#!/usr/bin/env python3

from .environ import env
from .output import ansi, output_iter, prints, get_ansi_len, remove_ansi

import torch
import torch.distributed as dist
from collections import defaultdict, deque
import datetime
import time


from typing import Generator, Iterable, TypeVar    # TODO: python 3.10
_T = TypeVar("_T")


class SmoothedValue(object):
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size: int = None, fmt: str = '{global_avg:.3f}'):
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value: float, n: int = 1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    def update_list(self, value_list: list[float]):
        for value in value_list:
            self.deque.append(value)
            self.total += value
        self.count += len(value_list)

    def reset(self):
        self.deque = deque(maxlen=self.deque.maxlen)
        self.count = 0
        self.total = 0.0

    def synchronize_between_processes(self):
        """
        Warning: does not synchronize the deque!
        """
        if not is_dist_avail_and_initialized():
            return
        t = torch.tensor([self.count, self.total], dtype=torch.float64, device='cuda')
        dist.barrier()
        dist.all_reduce(t)
        t = t.tolist()
        self.count = int(t[0])
        self.total = float(t[1])

    @property
    def median(self) -> float:
        try:
            d = torch.tensor(list(self.deque))
            return d.median().item()
        except Exception:
            return 0.0

    @property
    def avg(self) -> float:
        try:
            d = torch.tensor(list(self.deque), dtype=torch.float32)
            if len(d) == 0:
                return 0.0
            return d.mean().item()
        except Exception:
            return 0.0

    @property
    def global_avg(self) -> float:
        try:
            return self.total / self.count
        except Exception:
            return 0.0

    @property
    def max(self) -> float:
        try:
            return max(self.deque)
        except Exception:
            return 0.0

    @property
    def min(self) -> float:
        try:
            return min(self.deque)
        except Exception:
            return 0.0

    @property
    def value(self) -> float:
        try:
            return self.deque[-1]
        except Exception:
            return 0.0

    def __str__(self):
        return self.fmt.format(
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            min=self.min,
            max=self.max,
            value=self.value)

    def __format__(self, format_spec: str) -> str:
        return self.__str__()

    def __repr__(self) -> str:
        return self.__str__()


class MetricLogger(object):
    def __init__(self, delimiter: str = '', meter_length: int = 20, indent: int = 0):
        self.meters: defaultdict[str, SmoothedValue] = defaultdict(SmoothedValue)
        self.delimiter = delimiter
        self.meter_length = meter_length
        self.indent = indent

    def update(self, **kwargs):
        for k, v in kwargs.items():
            self.meters[k].update(float(v))

    def __getattr__(self, attr: str) -> float:
        if attr in self.meters:
            return self.meters[attr]
        if attr in self.__dict__:   # TODO: use hasattr
            return self.__dict__[attr]
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, attr))

    def __str__(self):
        loss_str: list[str] = []
        for name, meter in self.meters.items():
            _str = '{green}{}{reset}: {}'.format(name, str(meter), **ansi)
            max_length = self.meter_length + get_ansi_len(_str)
            if len(_str) > max_length:
                _str = '{green}{}{reset}: {}'.format(name, str(meter)[:5], **ansi)
            _str = _str.ljust(max_length)
            loss_str.append(_str)
        return self.delimiter.join(loss_str)

    def synchronize_between_processes(self):
        for meter in self.meters.values():
            meter.synchronize_between_processes()

    def log_every(self, iterable: Iterable[_T], header: str = None,
                  total: int = None, print_freq: int = 0,
                  indent: int = None) -> Generator[_T, None, None]:
        indent = indent if indent is not None else self.indent
        if total is None:
            try:
                total = len(iterable)
            except Exception:
                pass
        i = 0
        if not header:
            header = ''
        iter_time = SmoothedValue(fmt='{avg:.4f}')
        data_time = SmoothedValue(fmt='{avg:.4f}')
        memory = SmoothedValue(fmt='{max:.0f}')  # Memory is measured by Max value
        MB = 1024.0 * 1024.0

        end = time.time()
        start_time = time.time()
        for i, obj in enumerate(iterable):
            data_time.update(time.time() - end)
            yield obj
            iter_time.update(time.time() - end)
            if torch.cuda.is_available():
                memory.update(torch.cuda.max_memory_allocated() / MB)
            if print_freq and i % print_freq == 0:
                middle_header = '' if total is None else output_iter(i, total)
                length = max(len(remove_ansi(header)) - 10, 0)
                middle_header = middle_header.ljust(length + get_ansi_len(middle_header))
                log_msg = self.delimiter.join([middle_header, str(self)])
                if env['verbose'] > 1:
                    iter_time_str = '{green}iter{reset}: {iter_time} s'.format(iter_time=str(iter_time), **ansi)
                    data_time_str = '{green}data{reset}: {data_time} s'.format(data_time=str(data_time), **ansi)
                    iter_time_str = iter_time_str.ljust(self.meter_length + get_ansi_len(iter_time_str))
                    data_time_str = data_time_str.ljust(self.meter_length + get_ansi_len(data_time_str))
                    log_msg = self.delimiter.join([log_msg, iter_time_str, data_time_str])
                if env['verbose'] > 2 and torch.cuda.is_available():
                    memory_str = '{green}memory{reset}: {memory} MB'.format(memory=str(memory), **ansi)
                    memory_str = memory_str.ljust(self.meter_length + get_ansi_len(memory_str))
                    log_msg = self.delimiter.join([log_msg, memory_str])
                prints(log_msg, indent=indent + 10)
            end = time.time()
        self.synchronize_between_processes()
        total_time = time.time() - start_time
        total_time = str(datetime.timedelta(seconds=int(total_time)))

        total_time_str = '{green}time{reset}: {time}'.format(time=total_time, **ansi)
        total_time_str = total_time_str.ljust(self.meter_length + get_ansi_len(total_time_str))
        log_msg = self.delimiter.join([header, str(self), total_time_str])
        if env['verbose'] > 1:
            iter_time_str = '{green}iter{reset}: {iter_time} s'.format(iter_time=str(iter_time), **ansi)
            data_time_str = '{green}data{reset}: {data_time} s'.format(data_time=str(data_time), **ansi)
            iter_time_str = iter_time_str.ljust(self.meter_length + get_ansi_len(iter_time_str))
            data_time_str = data_time_str.ljust(self.meter_length + get_ansi_len(data_time_str))
            log_msg = self.delimiter.join([log_msg, iter_time_str, data_time_str])
        if env['verbose'] > 2 and torch.cuda.is_available():
            memory_str = '{green}memory{reset}: {memory} MB'.format(memory=str(memory), **ansi)
            memory_str = memory_str.ljust(self.meter_length + get_ansi_len(memory_str))
            log_msg = self.delimiter.join([log_msg, memory_str])
        prints(log_msg, indent=indent)


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True



#!/usr/bin/env python3

from autovul.base.environ import env
import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import TYPE_CHECKING
from autovul.base.models import Model    # TODO: python 3.10
if TYPE_CHECKING:
    import torch.utils.data


class InfluenceFunction():
    def __init__(self, model: Model, sample_size: int = 64, **kwargs):
        self.sample_size = sample_size
        self.model = model
        self.dataset = self.model.dataset
        self.module = self.get_module()
        self.parameter = self.get_parameter()
        self.criterion_vec = nn.CrossEntropyLoss(weight=self.dataset.loss_weights, reduction='none')

    def get_module(self) -> nn.Linear:
        return self.model._model.classifier[-1]

    def get_parameter(self) -> nn.Parameter:
        return self.module.weight

    def up_loss(self, z: torch.Tensor = None, z_label: torch.Tensor = None, v: torch.Tensor = None,
                z_test: torch.Tensor = None, z_test_label: torch.Tensor = None, v_test: torch.Tensor = None,
                hess_inv: torch.Tensor = None) -> list[float]:
        if v is None:
            v = self.calc_v(z, z_label)
        if (z_test is None and z_test_label is None):
            z_test = z
            z_test_label = z_label
        if v_test is None:
            if (z is None and z_label is None) or (z is z_test and z_label is z_test_label):
                v_test = v
            else:
                v_test = self.calc_v(z_test, z_test_label)
        s_test = v_test @ hess_inv
        return (v * s_test).sum(dim=1).detach().cpu().tolist()

    def calc_v(self, z: torch.Tensor, z_test_label: torch.Tensor) -> torch.Tensor:
        def func(weight: nn.Parameter) -> torch.Tensor:
            del self.module.weight
            self.module.weight = weight
            _output = self.model(_input=z)
            return self.criterion_vec(_output, z_test_label)
        jocobian: torch.Tensor = torch.autograd.functional.jacobian(func, self.parameter)
        return jocobian.flatten(1)

    # def calc_s_test(self, v_test: torch.Tensor, hess_inv: torch.Tensor = None) -> torch.Tensor:
    #     if isinstance(hess_inv, torch.Tensor):
    #         return v_test @ hess_inv

    def calc_H(self, loader: torch.utils.data.DataLoader = None, eps=1e-5, **kwargs) -> torch.Tensor:
        loader = loader if loader is not None else self.dataset.loader['train']
        hess_list: list[torch.Tensor] = []
        batch_num = max(self.sample_size // loader.batch_size, 1)
        for i, data in enumerate(loader):
            if i >= batch_num:
                break
            _input, _label = self.model.get_data(data, **kwargs)
            _feats = self.model.get_final_fm(_input)    # TODO
            hess = self.compute_hess(_feats, _label).detach().cpu()  # (N, D, D)
            hess_list.append(hess)
        hess = torch.stack(hess_list).mean(dim=0).to(env['device'])  # (D, D)
        return (hess + hess.t()) / 2 + eps * torch.eye(len(hess), device=hess.device)   # numerical robust

    def compute_fim(self, _feats: torch.Tensor) -> torch.Tensor:
        # _feats: (N, D)
        def func(weight: nn.Parameter) -> torch.Tensor:
            del self.module.weight  # (D, class_num)
            self.module.weight = weight
            log_prob = F.log_softmax(self.module(_feats))  # (N, class_num)
            return log_prob.mean(dim=0)  # (class_num)
        jacobian: torch.Tensor = torch.autograd.functional.jacobian(func, self.parameter)
        jacobian = jacobian.flatten(start_dim=2)  # (class_num, D*class_num)
        prob = F.softmax(self.module(_feats)).unsqueeze(-1).unsqueeze(-1)  # (N, class_num, 1, 1)
        hess = prob * jacobian.unsqueeze(-1) * jacobian.unsqueeze(-2)  # (N, class_num, D*class_num, D*class_num)
        hess = hess.sum(dim=1)  # (N, D*class_num, D*class_num)
        return hess

    def compute_hess(self, _feats: torch.Tensor, _label: torch.Tensor) -> torch.Tensor:
        # _feats: (N, D)
        def func(weight: nn.Parameter) -> torch.Tensor:
            del self.module.weight
            self.module.weight = weight
            return self.model.criterion(self.module(_feats), _label)
        hess: torch.Tensor = torch.autograd.functional.hessian(func, self.parameter)
        return hess.flatten(-2, -1).flatten(0, 1)
        # def func(weight):
        #     del self.module.weight
        #     self.module.weight = weight
        #     return self.model.criterion(self.module(_feats), _label)
        # return torch.autograd.functional.hessian(func, self.parameter).flatten(-2, -1).flatten(0, 1)

    # def calc_Ht(self, t: torch.Tensor, _input: torch.Tensor = None, _label: torch.Tensor = None) -> torch.Tensor:
    #     if _input is None and _label is None:
    #         dataset = self.dataset.loader['train'].dataset
    #         data = sample_batch(dataset, self.sample_size)
    #         _input, _label = self.model.get_data(data)
    #     loss = self.model.loss(_input=_input, _label=_label)

    #     self.parameter.requires_grad()
    #     grad = torch.autograd.grad(loss, self.parameter,
    #                                create_graph=True)[0].flatten()    # (D)
    #     Ht = torch.autograd.grad(grad @ t, self.parameter)[0].flatten().detach()
    #     self.parameter.requires_grad(False)
    #     return Ht    # (D)



#!/usr/bin/env python3

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _pair

from torch.types import _int, _size
from typing import Optional, Union


def weight_init(m: nn.Module, filter_list: tuple[type] = ()) -> None:
    # Function for Initialization
    '''
    Usage:
        model = Model()
        model.apply(weight_init)
    '''
    if isinstance(m, filter_list):
        return
    if hasattr(m, 'reset_parameters'):
        return m.reset_parameters()
    else:
        for layer in m.children():
            weight_init(layer, filter_list=filter_list)


def conv2d_same_padding(input: torch.Tensor, weight: torch.Tensor,
                        bias: Optional[torch.Tensor] = None,
                        stride: Union[_int, _size] = 1,
                        padding: Union[_int, _size] = 1,    # useless
                        dilation: Union[_int, _size] = 1,
                        groups: _int = 1):
    input_rows = input.size(2)
    filter_rows = weight.size(2)
    # effective_filter_size_rows = (filter_rows - 1) * dilation[0] + 1
    out_rows = (input_rows + stride[0] - 1) // stride[0]
    padding_rows = max(0, (out_rows - 1) * stride[0] +
                       (filter_rows - 1) * dilation[0] + 1 - input_rows)
    rows_odd = (padding_rows % 2 != 0)
    padding_cols = max(0, (out_rows - 1) * stride[0] +
                       (filter_rows - 1) * dilation[0] + 1 - input_rows)
    cols_odd = (padding_rows % 2 != 0)

    if rows_odd or cols_odd:
        input = F.pad(input, [0, int(cols_odd), 0, int(rows_odd)])

    return F.conv2d(input, weight, bias, stride[0],
                    padding=(padding_rows // 2, padding_cols // 2),
                    dilation=dilation, groups=groups)


class Conv2d_SAME(_ConvNd):
    def __init__(self, in_channels: int, out_channels: int,
                 kernel_size: tuple[int, ...], stride: tuple[int, ...] = 1,
                 padding: tuple[int, ...] = 0, dilation: tuple[int, ...] = 1,
                 groups: int = 1, bias: bool = True, padding_mode: str = 'zeros'):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            transposed=False, output_padding=_pair(0),
            groups=groups, bias=bias, padding_mode=padding_mode)

    def forward(self, input):
        return conv2d_same_padding(input, self.weight, self.bias, self.stride,
                                   self.padding, self.dilation, self.groups)



#!/usr/bin/env python3

from autovul.vision.datasets.imageset import ImageSet
from autovul.vision.utils.dataset import ZipFolder
from autovul.vision.environ import env
from autovul.base.utils.output import ansi, prints, output_iter

import torchvision.datasets as datasets
from torchvision.datasets.utils import check_integrity, download_and_extract_archive, extract_archive
import numpy as np
import json
import zipfile
import os
import glob
import shutil
from tqdm import tqdm

from typing import Union
from typing import TYPE_CHECKING
import argparse    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class ImageFolder(ImageSet):

    name: str = 'imagefolder'
    url: dict[str, str] = {}
    ext = {'train': '.zip', 'valid': '.zip', 'test': '.zip'}    # TODO: Use Param?
    md5: dict[str, str] = {}
    org_folder_name: dict[str, str] = {}

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--data_format', choices=['folder', 'tar', 'zip'],
                           help='file format of dataset. (zip is using ZIP_STORED)')
        group.add_argument('--memory', action='store_true',
                           help='put all dataset into memory initialization.')
        return group

    def __init__(self, data_format: str = 'folder', memory: bool = False, **kwargs):
        self.data_format: str = data_format
        self.memory: bool = memory
        super().__init__(**kwargs)
        self.param_list['imagefolder'] = ['data_format', 'url', 'org_folder_name', 'memory']
        self.class_to_idx = self.get_class_to_idx()
        if self.num_classes is None:
            self.num_classes = len(self.class_to_idx)

    def initialize(self, *args, **kwargs):
        if self.data_format == 'folder' or not self.check_files(data_format='folder'):
            self.initialize_folder(*args, **kwargs)
        if self.data_format == 'zip':
            self.initialize_zip(*args, **kwargs)

    def initialize_folder(self, **kwargs):
        print('initialize folder')
        mode_list = ['train', 'valid'] if self.valid_set and 'valid' in self.url.keys() else ['train']
        for mode in mode_list:
            zip_path = os.path.join(self.folder_path, f'{self.name}_{mode}_store.zip')
            if os.path.isfile(zip_path):
                print('{yellow}Uncompress file{reset}: '.format(**ansi), zip_path)
                extract_archive(from_path=zip_path, to_path=self.folder_path)
                print('{green}Uncompress finished{reset}: '.format(**ansi),
                      f'{zip_path}')
                print()
                continue
            tar_path = os.path.join(self.folder_path, f'{self.name}_{mode}_store.zip')
            self.download_and_extract_archive(mode=mode)
            os.rename(os.path.join(self.folder_path, self.org_folder_name[mode]),
                      os.path.join(self.folder_path, mode))
            try:
                dirname = os.path.dirname(self.org_folder_name[mode])
                if dirname:
                    shutil.rmtree(os.path.join(self.folder_path, dirname))
            except FileNotFoundError:
                pass

    def download_and_extract_archive(self, mode: str):
        file_name = f'{self.name}_{mode}{self.ext[mode]}'
        file_path = os.path.normpath(os.path.join(self.folder_path, file_name))
        md5 = None if mode not in self.md5.keys() else self.md5[mode]
        if not check_integrity(file_path, md5=md5):
            prints('{yellow}Downloading Dataset{reset} '.format(**ansi),
                   f'{self.name} {mode:5s}: {file_path}', indent=10)
            download_and_extract_archive(url=self.url[mode],
                                         download_root=self.folder_path, extract_root=self.folder_path,
                                         filename=file_name, md5=md5)
            prints('{upline}{clear_line}'.format(**ansi), indent=10)
        else:
            prints('{yellow}File Already Exists{reset}: '.format(**ansi), file_path, indent=10)
            extract_archive(from_path=file_path, to_path=self.folder_path)

    def initialize_zip(self, mode_list: list[str] = None, **kwargs):
        if mode_list is None:
            mode_list = [mode for mode in ['train', 'valid', 'test']
                         if os.path.isdir(os.path.join(self.folder_path, mode))]
        for mode in mode_list:
            src_path = os.path.normpath(os.path.join(self.folder_path, mode))
            dst_path = os.path.join(self.folder_path, f'{self.name}_{mode}_store.zip')
            assert os.path.isdir(src_path)
            if not os.path.exists(dst_path):
                print('{yellow}initialize zip{reset}: '.format(**ansi), dst_path)
                ZipFolder.initialize_from_folder(root=src_path, zip_path=dst_path)
                print('{green}initialize zip finish{reset}'.format(**ansi))

    def _get_org_dataset(self, mode: str, data_format: str = None, **kwargs) -> datasets.DatasetFolder:
        data_format = self.data_format if data_format is None else data_format
        root = os.path.join(self.folder_path, mode)
        DatasetClass = datasets.ImageFolder
        if data_format == 'zip':
            root = os.path.join(self.folder_path, f'{self.name}_{mode}_store.zip')
            DatasetClass = ZipFolder
            if 'memory' not in kwargs.keys():
                kwargs['memory'] = self.memory
        return DatasetClass(root=root, **kwargs)

    def get_class_to_idx(self, file_path: str = None) -> dict[str, int]:
        if file_path is None:
            file_path = os.path.join(self.folder_path, 'class_to_idx.json')
        if os.path.exists(file_path):
            with open(file_path) as fp:
                result: dict[str, int] = json.load(fp)
            return result
        return self.get_org_dataset('train').class_to_idx

    def sample(self, child_name: str = None, class_dict: dict[str, list[str]] = None, sample_num: int = None,
               method='zip'):
        if sample_num is None:
            assert class_dict
            sample_num = len(class_dict)
        if child_name is None:
            child_name = self.name + '_sample%d' % sample_num
        src_path = self.folder_path
        dst_path = os.path.normpath(os.path.join(os.path.dirname(self.folder_path), child_name))
        if not os.path.exists(dst_path):
            os.makedirs(dst_path)
        print('{yellow}src path{reset}: '.format(**ansi), src_path)
        print('{yellow}dst path{reset}: '.format(**ansi), dst_path)

        mode_list = [mode for mode in ['train', 'valid', 'test'] if os.path.isdir(os.path.join(src_path, mode))]
        if method == 'zip':
            zip_path_list: list[str] = glob.glob(os.path.join(src_path, '*_store.zip'))
            mode_list = [os.path.basename(zip_path).removeprefix(self.name).removesuffix('_store.zip')
                         for zip_path in zip_path_list]

        src2dst_dict: dict[str, str] = {}
        if class_dict is None:
            assert sample_num
            idx_list = np.arange(self.num_classes)
            np.random.seed(env['data_seed'])
            np.random.shuffle(idx_list)
            idx_list = idx_list[:sample_num]
            mode = mode_list[0]
            class_list: list[str] = []
            if method == 'zip':
                zip_path = os.path.join(src_path, f'{self.name}_{mode}_store.zip')
                with zipfile.ZipFile(zip_path, 'r', compression=zipfile.ZIP_STORED) as src_zip:
                    name_list = src_zip.namelist()
                for name in name_list:
                    name_dir, name_base = os.path.split(os.path.dirname(name))
                    if name_dir == mode:
                        class_list.append(name_base)
            elif method == 'folder':
                folder_path = os.path.join(src_path, f'{mode}')
                class_list = np.array(os.listdir(folder_path))[idx_list].tolist()
                class_list = [_dir for _dir in class_list if os.path.isdir(os.path.join(folder_path, _dir))]
            class_list.sort()
            class_list = np.array(class_list)[idx_list].tolist()
            for class_name in class_list:
                src2dst_dict[class_name] = class_name
        else:
            src2dst_dict = {src_class: dst_class for src_class, dst_list in class_dict.items()
                            for dst_class in dst_list}
        src_class_list = src2dst_dict.keys()
        print(src2dst_dict)
        if method == 'zip':
            for mode in mode_list:
                print('{purple}mode: {0}{reset}'.format(mode, **ansi))
                assert mode in ['train', 'valid', 'test']
                dst_zip = zipfile.ZipFile(os.path.join(dst_path, f'{child_name}_{mode}_store.zip'),
                                          'w', compression=zipfile.ZIP_STORED)
                src_zip = zipfile.ZipFile(os.path.join(src_path, f'{self.name}_{mode}_store.zip'),
                                          'r', compression=zipfile.ZIP_STORED)
                _list = src_zip.namelist()
                if env['tqdm']:
                    _list = tqdm(_list)
                for filename in _list:
                    if filename[-1] == '/':
                        continue
                    dirname, basename = os.path.split(filename)
                    mode_check, src_class = os.path.split(dirname)
                    if mode_check == mode and src_class in src_class_list:
                        print(filename)
                        dst_class = src2dst_dict[src_class]
                        dst_zip.writestr(f'{mode}/{dst_class}/{basename}',
                                         src_zip.read(filename))
                src_zip.close()
                dst_zip.close()
        elif method == 'folder':
            len_i = len(class_dict.keys())
            for mode in mode_list:
                print('{purple}{0}{reset}'.format(mode, **ansi))
                assert mode in ['train', 'valid', 'test']
                for i, dst_class in enumerate(class_dict.keys()):
                    if not os.path.exists(_path := os.path.join(dst_path, mode, dst_class)):
                        os.makedirs(_path)
                    prints('{blue_light}{0}{reset}'.format(dst_class, **ansi), indent=10)
                    class_list = class_dict[dst_class]
                    len_j = len(class_list)
                    for j, src_class in enumerate(class_list):
                        _list = os.listdir(os.path.join(src_path, mode, src_class))
                        prints(output_iter(i + 1, len_i) + output_iter(j + 1, len_j) +
                               f'dst: {dst_class:15s}    src: {src_class:15s}    image_num: {len(_list):>8d}', indent=10)
                        if env['tqdm']:
                            _list = tqdm(_list)
                        for _file in _list:
                            shutil.copyfile(os.path.join(src_path, mode, src_class, _file),
                                            os.path.join(dst_path, mode, dst_class, _file))
                        if env['tqdm']:
                            print('{upline}{clear_line}'.format(**ansi))



#!/usr/bin/env python3

from torchvision import get_image_backend
from torchvision.datasets import VisionDataset, DatasetFolder
from torchvision.datasets.folder import has_file_allowed_extension, IMG_EXTENSIONS
import numpy as np
import PIL.Image as Image
import io
import os
import tarfile
import zipfile
from typing import Any, cast
from collections.abc import Callable    # TODO: python 3.10
from typing import Optional, Union  # TODO: python 3.10


__all__ = ['MemoryDataset', 'ZipFolder']


# TODO: Need reorganization
class MemoryDataset(VisionDataset):
    def __init__(self, data: np.ndarray = None, targets: list[int] = None,
                 root: str = None, **kwargs):
        super().__init__(root=root, **kwargs)
        self.data = data
        self.targets = targets
        if data is None and os.path.isfile(root) and root.endswith('.npz'):
            _dict = np.load(root)
            self.data = _dict['data']
            self.targets = list(_dict['targets'])
        assert len(self.data) == len(self.targets)

    def __getitem__(self, index: Union[int, slice]) -> tuple[Any, Any]:
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]
        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def __len__(self):
        return len(self.targets)


# https://github.com/koenvandesande/vision/blob/read_zipped_data/torchvision/datasets/zippedfolder.py
class ZipFolder(DatasetFolder):
    def __init__(self, root: str, transform: Optional[Callable] = None, target_transform: Optional[Callable] = None,
                 is_valid_file: Optional[Callable[[str], bool]] = None, memory: bool = True) -> None:
        if not root.lower().endswith('.zip'):
            raise TypeError("Need zip file for data source: ", root)
        if memory:
            with open(root, 'rb') as zf:
                data = zf.read()
            self.root_data = zipfile.ZipFile(io.BytesIO(data), 'r')
        else:
            self.root_data = zipfile.ZipFile(root, 'r')
        super().__init__(root, self.zip_loader, IMG_EXTENSIONS if is_valid_file is None else None,
                         transform=transform, target_transform=target_transform, is_valid_file=is_valid_file)
        self.imgs = self.samples

    @staticmethod
    def initialize_from_folder(root: str, zip_path: str = None):
        root = os.path.normpath(root)
        folder_dir, folder_base = os.path.split(root)
        if zip_path is None:
            zip_path = os.path.join(folder_dir, f'{folder_base}_store.zip')
        with zipfile.ZipFile(zip_path, mode='w', compression=zipfile.ZIP_STORED) as zf:
            for walk_root, walk_dirs, walk_files in os.walk(root):
                zip_root = walk_root.removeprefix(folder_dir)
                for _file in walk_files:
                    org_path = os.path.join(walk_root, _file)
                    dst_path = os.path.join(zip_root, _file)
                    zf.write(org_path, dst_path)

    def make_dataset(
        self,
        directory: str,
        class_to_idx: dict[str, int],
        extensions: Optional[tuple[str, ...]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ) -> list[tuple[str, int]]:
        instances = []
        both_none = extensions is None and is_valid_file is None
        both_something = extensions is not None and is_valid_file is not None
        if both_none or both_something:
            raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")
        if extensions is not None:
            def is_valid_file(x: str) -> bool:
                return has_file_allowed_extension(x, cast(tuple[str, ...], extensions))
        is_valid_file = cast(Callable[[str], bool], is_valid_file)
        for filepath in self.root_data.namelist():
            if is_valid_file(filepath):
                target_class = os.path.basename(os.path.dirname(filepath))
                instances.append((filepath, class_to_idx[target_class]))
        return instances

    def zip_loader(self, path: str) -> Image.Image:
        f = self.root_data.open(path, 'r')
        if get_image_backend() == 'accimage':
            try:
                import accimage  # type: ignore
                return accimage.Image(f)
            except IOError:
                pass   # fall through to PIL
        return Image.open(f).convert('RGB')

    def _find_classes(self, *args, **kwargs) -> tuple[list[str], dict[str, int]]:
        """
        Finds the class folders in a dataset.
        Args:
            dir (string): Root directory path.
        Returns:
            tuple: (classes, class_to_idx) where classes are relative to (dir), and class_to_idx is a dictionary.
        Ensures:
            No class is a subdirectory of another.
        """
        classes: set[str] = set()
        for filepath in self.root_data.namelist():
            root, target_class = os.path.split(os.path.dirname(filepath))
            if root:
                classes.add(target_class)
        classes = list(classes)
        classes.sort()  # TODO: Pylance issue
        class_to_idx = {classes[i]: i for i in range(len(classes))}
        return classes, class_to_idx


class TarFolder(DatasetFolder):
    def __init__(self, root: str, transform: Optional[Callable] = None, target_transform: Optional[Callable] = None,
                 is_valid_file: Optional[Callable[[str], bool]] = None, memory: bool = True) -> None:
        if not root.lower().endswith('.tar'):
            raise TypeError("Need tar file for data source: ", root)
        if memory:
            with open(root, 'rb') as zf:
                data = zf.read()
            self.root_data = tarfile.open(io.BytesIO(data), 'r')
        else:
            self.root_data = tarfile.open(root, 'r')
        super().__init__(root, self.tar_loader, IMG_EXTENSIONS if is_valid_file is None else None,
                         transform=transform, target_transform=target_transform, is_valid_file=is_valid_file)
        self.imgs = self.samples

    @staticmethod
    def initialize_from_folder(root: str, tar_path: str = None) -> None:
        root = os.path.normpath(root)
        folder_dir, folder_base = os.path.split(root)
        if tar_path is None:
            tar_path = os.path.join(folder_dir, f'{folder_base}.tar')
        with tarfile.open(tar_path, mode='w') as zf:
            for walk_root, walk_dirs, walk_files in os.walk(root):
                tar_root = walk_root.removeprefix(folder_dir)
                for _file in walk_files:
                    org_path = os.path.join(walk_root, _file)
                    dst_path = os.path.join(tar_root, _file)
                    zf.add(org_path, dst_path)

    def make_dataset(
        self,
        directory: str,
        class_to_idx: dict[str, int],
        extensions: Optional[tuple[str, ...]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ) -> list[tuple[str, int]]:
        instances: list[tuple[str, int]] = []
        both_none = extensions is None and is_valid_file is None
        both_something = extensions is not None and is_valid_file is not None
        if both_none or both_something:
            raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")
        if extensions is not None:
            def is_valid_file(x: str) -> bool:
                return has_file_allowed_extension(x, cast(tuple[str, ...], extensions))
        is_valid_file = cast(Callable[[str], bool], is_valid_file)
        for filepath in self.root_data.getnames():
            if is_valid_file(filepath):
                target_class = os.path.basename(os.path.dirname(filepath))
                instances.append((filepath, class_to_idx[target_class]))
        return instances

    def tar_loader(self, path: str) -> Image.Image:
        f = self.root_data.extract(path)
        if get_image_backend() == 'accimage':
            try:
                import accimage  # type: ignore
                return accimage.Image(f)
            except IOError:
                pass   # fall through to PIL
        return Image.open(f).convert('RGB')

    def _find_classes(self, *args, **kwargs) -> tuple[list[str], dict[str, int]]:
        """
        Finds the class folders in a dataset.
        Args:
            dir (string): Root directory path.
        Returns:
            tuple: (classes, class_to_idx) where classes are relative to (dir), and class_to_idx is a dictionary.
        Ensures:
            No class is a subdirectory of another.
        """
        classes: set[str] = set()
        for filepath in self.root_data.getnames():
            root, target_class = os.path.split(os.path.dirname(filepath))
            if root:
                classes.add(target_class)
        classes = list(classes)
        classes.sort()  # TODO: Pylance issue
        class_to_idx = {classes[i]: i for i in range(len(classes))}
        return classes, class_to_idx



#!/usr/bin/env python3

# CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --epoch 200 --batch_size 96 --cutout --grad_clip 5.0 --lr 0.025 --lr_scheduler --save --dataset cifar10 --model resnet18_comp

# CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --dataset cifar10 --adv_train --adv_train_random_init --validate_interval 1 --epoch 15 --lr 0.1 --lr_scheduler --model resnet18_comp

# CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --dataset cifar10 --adv_train --adv_train_random_init --adv_train_iter 1 --adv_train_alpha 0.0392156862745 --adv_train_eval_iter 7 --adv_train_eval_alpha 0.0078431372549 --validate_interval 1 --epoch 15 --lr 0.1 --lr_scheduler --model resnet18_comp

# CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --dataset mnist --adv_train --adv_train_random_init --adv_train_iter 1 --adv_train_alpha 0.375 --adv_train_eps 0.3 --adv_train_eval_iter 7 --adv_train_eval_alpha 0.1 --adv_train_eval_eps 0.3 --validate_interval 1 --epoch 15 --lr 0.1 --lr_scheduler --model net

import autovul.vision
from autovul.vision.utils import summary
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    autovul.vision.environ.add_argument(parser)
    autovul.vision.datasets.add_argument(parser)
    autovul.vision.models.add_argument(parser)
    autovul.vision.trainer.add_argument(parser)
    args = parser.parse_args()

    env = autovul.vision.environ.create(**args.__dict__)
    dataset = autovul.vision.datasets.create(**args.__dict__)
    model = autovul.vision.models.create(dataset=dataset, **args.__dict__)
    trainer = autovul.vision.trainer.create(dataset=dataset, model=model, **args.__dict__)

    if env['verbose']:
        summary(env=env, dataset=dataset, model=model, trainer=trainer)
    model._train(**trainer)



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

from autovul.vision.utils.model_archs import dpn

import torch.nn as nn
from collections import OrderedDict
from collections.abc import Callable


class _DPN(_ImageModel):

    def __init__(self, name: str = 'dpn92', **kwargs):
        super().__init__(**kwargs)
        ModelClass: Callable[..., dpn.DPN] = getattr(dpn, name.replace('_comp', ''))
        _model = ModelClass(num_classes=self.num_classes)
        module_list: list[nn.Module] = []
        if 'comp' in name:
            conv1: nn.Conv2d = _model.conv1
            _model.conv1 = nn.Conv2d(conv1.in_channels, conv1.out_channels,
                                     kernel_size=3, stride=1, padding=1, bias=False)
            module_list.extend([
                ('conv1', _model.conv1),
                ('bn1', _model.bn1),
                ('relu', nn.ReLU(inplace=True)),
            ])
        else:
            module_list.extend([
                ('conv1', _model.conv1),
                ('bn1', _model.bn1),
                ('relu', nn.ReLU(inplace=True)),
                ('maxpool', _model.maxpool),
            ])
        module_list.extend([
            ('layer1', _model.layer1),
            ('layer2', _model.layer2),
            ('layer3', _model.layer3),
            ('layer4', _model.layer4),
        ])
        self.features = nn.Sequential(OrderedDict(module_list))
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', _model.fc)
        ]))


class DPN(ImageModel):
    available_models = ['dpn', 'dpn_comp',
                        'dpn68', 'dpn92', 'dpn98', 'dpn131', 'dpn107',
                        'dpn68_comp', 'dpn92_comp', 'dpn98_comp', 'dpn131_comp', 'dpn107_comp']

    def __init__(self, name: str = 'dpn', layer: int = 92,
                 model: type[_DPN] = _DPN, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch
import torch.nn as nn
import torchvision.models
from torchvision.models.densenet import model_urls as urls
import re
from collections import OrderedDict


class _DenseNet(_ImageModel):

    def __init__(self, name: str = 'densenet121', **kwargs):
        super().__init__(**kwargs)
        ModelClass: type[torchvision.models.DenseNet] = getattr(torchvision.models, name.split('_')[0])
        _model = ModelClass(num_classes=self.num_classes)
        self.features = _model.features
        self.features.add_module('relu', nn.ReLU(inplace=True))
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', _model.classifier)  # nn.Linear(512 * block.expansion, num_classes)
        ]))
        if 'comp' in name:
            conv0: nn.Conv2d = self.features.conv0
            conv0 = nn.Conv2d(conv0.in_channels, conv0.out_channels,
                              kernel_size=3, padding=1, bias=False)
            module_list = [('conv0', conv0)]
            module_list.extend(list(self.features.named_children())[4:])
            self.features = nn.Sequential(OrderedDict(module_list))


class DenseNet(ImageModel):
    available_models = ['densenet', 'densenet_comp',
                        'densenet121', 'densenet169', 'densenet201', 'densenet161',
                        'densenet121_comp', 'densenet169_comp', 'densenet201_comp', 'densenet161_comp']
    model_urls = urls

    def __init__(self, name: str = 'densenet', layer: int = 121,
                 model: type[_DenseNet] = _DenseNet, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        _dict = super().get_official_weights(**kwargs)
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        for key in list(_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                _dict[new_key] = _dict[key]
                del _dict[key]
        _dict['classifier.fc.weight'] = _dict['classifier.weight']
        _dict['classifier.fc.bias'] = _dict['classifier.bias']
        del _dict['classifier.weight']
        del _dict['classifier.bias']
        return _dict



#!/usr/bin/env python3

from collections import namedtuple

Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')

PRIMITIVES = [
    'max_pool_3x3',
    'avg_pool_3x3',
    'skip_connect',  # identity
    'sep_conv_3x3',
    'sep_conv_5x5',
    'dil_conv_3x3',
    'dil_conv_5x5',
    'none',  # zero
]

# https://github.com/quark0/darts
# https://github.com/automl/RobustDARTS
nasnet = Genotype(
    normal=[('sep_conv_5x5', 1), ('sep_conv_3x3', 0), ('sep_conv_5x5', 0), ('sep_conv_3x3', 0), ('avg_pool_3x3', 1),
            ('skip_connect', 0), ('avg_pool_3x3', 0), ('avg_pool_3x3', 0), ('sep_conv_3x3', 1), ('skip_connect', 1), ],
    normal_concat=[2, 3, 4, 5, 6],
    reduce=[('sep_conv_5x5', 1), ('sep_conv_7x7', 0), ('max_pool_3x3', 1), ('sep_conv_7x7', 0), ('avg_pool_3x3', 1),
            ('sep_conv_5x5', 0), ('skip_connect', 3), ('avg_pool_3x3', 2), ('sep_conv_3x3', 2), ('max_pool_3x3', 1), ],
    reduce_concat=[4, 5, 6],
)
nasnet_adapt = Genotype(
    normal=[('sep_conv_5x5', 1), ('sep_conv_3x3', 0), ('sep_conv_5x5', 1), ('sep_conv_3x3', 0), ('avg_pool_3x3', 1),
            ('skip_connect', 0), ('avg_pool_3x3', 1), ('avg_pool_3x3', 0), ('sep_conv_3x3', 1), ('skip_connect', 0), ],
    normal_concat=[2, 3, 4, 5, 6],
    reduce=[('sep_conv_5x5', 1), ('sep_conv_7x7', 0), ('max_pool_3x3', 1), ('sep_conv_7x7', 0), ('avg_pool_3x3', 1),
            ('sep_conv_5x5', 0), ('skip_connect', 1), ('avg_pool_3x3', 0), ('sep_conv_3x3', 1), ('max_pool_3x3', 0), ],
    reduce_concat=[4, 5, 6],
)

amoebanet = Genotype(
    normal=[('avg_pool_3x3', 0), ('max_pool_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_5x5', 2), ('sep_conv_3x3', 0),
            ('avg_pool_3x3', 3), ('sep_conv_3x3', 1), ('skip_connect', 1), ('skip_connect', 0), ('avg_pool_3x3', 1), ],
    normal_concat=[4, 5, 6],
    reduce=[('avg_pool_3x3', 0), ('sep_conv_3x3', 1), ('max_pool_3x3', 0), ('sep_conv_7x7', 2), ('sep_conv_7x7', 0),
            ('avg_pool_3x3', 1), ('max_pool_3x3', 0), ('max_pool_3x3', 1), ('conv_7x1_1x7', 0), ('sep_conv_3x3', 5), ],
    reduce_concat=[3, 4, 6]
)
amoebanet_adapt = Genotype(
    normal=[('avg_pool_3x3', 0), ('max_pool_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_5x5', 1), ('sep_conv_3x3', 0),
            ('avg_pool_3x3', 1), ('sep_conv_3x3', 0), ('skip_connect', 1), ('skip_connect', 0), ('avg_pool_3x3', 1), ],
    normal_concat=[4, 5, 6],
    reduce=[('avg_pool_3x3', 0), ('sep_conv_3x3', 1), ('max_pool_3x3', 0), ('sep_conv_7x7', 1), ('sep_conv_7x7', 0),
            ('avg_pool_3x3', 1), ('max_pool_3x3', 0), ('max_pool_3x3', 1), ('conv_7x1_1x7', 0), ('sep_conv_3x3', 1), ],
    reduce_concat=[3, 4, 6]
)

darts_v1 = Genotype(
    normal=[('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('skip_connect', 0), ('sep_conv_3x3', 1),
            ('skip_connect', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('skip_connect', 2)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 0),
            ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('avg_pool_3x3', 0)],
    reduce_concat=[2, 3, 4, 5])
darts_v2 = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 1),
            ('sep_conv_3x3', 1), ('skip_connect', 0), ('skip_connect', 0), ('dil_conv_3x3', 2)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1),
            ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('max_pool_3x3', 1)],
    reduce_concat=[2, 3, 4, 5])
darts = darts_v2

snas_mild = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('skip_connect', 0), ('dil_conv_3x3', 1),
            ('skip_connect', 0), ('skip_connect', 1), ('skip_connect', 0), ('sep_conv_3x3', 1)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1),
            ('max_pool_3x3', 1), ('skip_connect', 2), ('dil_conv_5x5', 2), ('max_pool_3x3', 0)],
    reduce_concat=[2, 3, 4, 5])

snas_adapt = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('skip_connect', 0), ('dil_conv_3x3', 1),
            ('skip_connect', 0), ('skip_connect', 1), ('skip_connect', 0), ('sep_conv_3x3', 1)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 0), ('max_pool_3x3', 1),
            ('max_pool_3x3', 0), ('skip_connect', 1), ('dil_conv_5x5', 0), ('max_pool_3x3', 1)],
    reduce_concat=[2, 3, 4, 5])

enas = Genotype(
    normal=[('sep_conv_3x3', 1), ('skip_connect', 1), ('sep_conv_5x5', 1), ('skip_connect', 0), ('avg_pool_3x3', 0),
            ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('avg_pool_3x3', 1), ('sep_conv_5x5', 1), ('avg_pool_3x3', 0)],
    normal_concat=[2, 3, 4, 5, 6],
    reduce=[('sep_conv_5x5', 0), ('avg_pool_3x3', 1), ('sep_conv_3x3', 1), ('avg_pool_3x3', 1), ('sep_conv_3x3', 1),
            ('avg_pool_3x3', 1), ('sep_conv_5x5', 4), ('avg_pool_3x3', 1), ('sep_conv_3x3', 5), ('sep_conv_5x5', 0)],
    reduce_concat=[2, 3, 6])

enas_adapt = Genotype(
    normal=[('sep_conv_3x3', 0), ('skip_connect', 1), ('sep_conv_5x5', 0), ('skip_connect', 0), ('avg_pool_3x3', 0),
            ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('avg_pool_3x3', 1), ('sep_conv_5x5', 0), ('avg_pool_3x3', 1)],
    normal_concat=[2, 3, 4, 5, 6],
    reduce=[('sep_conv_5x5', 0), ('avg_pool_3x3', 1), ('sep_conv_3x3', 0), ('avg_pool_3x3', 1), ('sep_conv_3x3', 0),
            ('avg_pool_3x3', 1), ('sep_conv_5x5', 0), ('avg_pool_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_5x5', 1)],
    reduce_concat=[2, 3, 6])

robust_darts = Genotype(
    normal=[('skip_connect', 1), ('dil_conv_3x3', 0), ('skip_connect', 2), ('skip_connect', 0),
            ('skip_connect', 0), ('skip_connect', 2), ('skip_connect', 2), ('skip_connect', 1)],
    normal_concat=range(2, 6),
    reduce=[('max_pool_3x3', 1), ('max_pool_3x3', 0), ('skip_connect', 2), ('max_pool_3x3', 1),
            ('skip_connect', 2), ('max_pool_3x3', 0), ('max_pool_3x3', 0), ('max_pool_3x3', 1)],
    reduce_concat=range(2, 6))

# https://github.com/chenxin061/pdarts
pdarts = Genotype(
    normal=[('skip_connect', 0), ('dil_conv_3x3', 1), ('skip_connect', 0), ('sep_conv_3x3', 1),
            ('sep_conv_3x3', 1), ('sep_conv_3x3', 3), ('sep_conv_3x3', 0), ('dil_conv_5x5', 4)],
    normal_concat=range(2, 6),
    reduce=[('avg_pool_3x3', 0), ('sep_conv_5x5', 1), ('sep_conv_3x3', 0), ('dil_conv_5x5', 2),
            ('max_pool_3x3', 0), ('dil_conv_3x3', 1), ('dil_conv_3x3', 1), ('dil_conv_5x5', 3)],
    reduce_concat=range(2, 6))

# https://github.com/yuhuixu1993/PC-DARTS
pc_darts_cifar = Genotype(
    normal=[('sep_conv_3x3', 1), ('skip_connect', 0), ('sep_conv_3x3', 0), ('dil_conv_3x3', 1),
            ('sep_conv_5x5', 0), ('sep_conv_3x3', 1), ('avg_pool_3x3', 0), ('dil_conv_3x3', 1)],
    normal_concat=range(2, 6),
    reduce=[('sep_conv_5x5', 1), ('max_pool_3x3', 0), ('sep_conv_5x5', 1), ('sep_conv_5x5', 2),
            ('sep_conv_3x3', 0), ('sep_conv_3x3', 3), ('sep_conv_3x3', 1), ('sep_conv_3x3', 2)],
    reduce_concat=range(2, 6))
pc_darts_image = Genotype(
    normal=[('skip_connect', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 0), ('skip_connect', 1),
            ('sep_conv_3x3', 1), ('sep_conv_3x3', 3), ('sep_conv_3x3', 1), ('dil_conv_5x5', 4)],
    normal_concat=range(2, 6),
    reduce=[('sep_conv_3x3', 0), ('skip_connect', 1), ('dil_conv_5x5', 2), ('max_pool_3x3', 1),
            ('sep_conv_3x3', 2), ('sep_conv_3x3', 1), ('sep_conv_5x5', 0), ('sep_conv_3x3', 3)],
    reduce_concat=range(2, 6))
pc_darts = pc_darts_cifar

# https://github.com/xiangning-chen/DrNAS
drnas_cifar10 = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_5x5', 1), ('sep_conv_3x3', 1), ('sep_conv_3x3', 2),
            ('skip_connect', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 2), ('dil_conv_5x5', 3)],
    normal_concat=range(2, 6),
    reduce=[('max_pool_3x3', 0), ('sep_conv_5x5', 1), ('dil_conv_5x5', 2), ('sep_conv_5x5', 1),
            ('sep_conv_5x5', 1), ('dil_conv_5x5', 3), ('skip_connect', 4), ('sep_conv_5x5', 1)],
    reduce_concat=range(2, 6))
drnas_imagenet = Genotype(
    normal=[('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 0), ('sep_conv_3x3', 1),
            ('sep_conv_3x3', 0), ('dil_conv_3x3', 3), ('skip_connect', 0), ('sep_conv_3x3', 1)],
    normal_concat=range(2, 6),
    reduce=[('max_pool_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('skip_connect', 2),
            ('sep_conv_3x3', 0), ('sep_conv_3x3', 2), ('sep_conv_3x3', 3), ('sep_conv_3x3', 1)],
    reduce_concat=range(2, 6))
drnas = drnas_cifar10

# https://github.com/lightaime/sgas
sgas = Genotype(
    normal=[('sep_conv_3x3', 0), ('dil_conv_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 1),
            ('sep_conv_5x5', 1), ('sep_conv_5x5', 3), ('skip_connect', 0), ('dil_conv_5x5', 2)],
    normal_concat=range(2, 6),
    reduce=[('max_pool_3x3', 0), ('dil_conv_5x5', 1), ('sep_conv_3x3', 0), ('skip_connect', 2),
            ('sep_conv_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 0), ('max_pool_3x3', 1)],
    reduce_concat=range(2, 6))


diy_deep = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 2), ('sep_conv_3x3', 2),
            ('sep_conv_3x3', 3), ('skip_connect', 3), ('skip_connect', 4), ('dil_conv_3x3', 4)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1),
            ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('max_pool_3x3', 1)],
    reduce_concat=[2, 3, 4, 5])

diy_noskip = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 1),
            ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('dil_conv_3x3', 0), ('dil_conv_3x3', 2)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1),
            ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('max_pool_3x3', 1)],
    reduce_concat=[2, 3, 4, 5])

diy_deep_noskip = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 2), ('sep_conv_3x3', 2),
            ('sep_conv_3x3', 3), ('sep_conv_3x3', 3), ('dil_conv_3x3', 4), ('dil_conv_3x3', 4)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1),
            ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('max_pool_3x3', 1)],
    reduce_concat=[2, 3, 4, 5])

random = Genotype(
    normal=[('max_pool_3x3', 0), ('dil_conv_3x3', 1), ('sep_conv_3x3', 2), ('skip_connect', 1),
            ('skip_connect', 3), ('avg_pool_3x3', 0), ('sep_conv_3x3', 1), ('dil_conv_3x3', 2)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('avg_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 1), ('dil_conv_3x3', 1),
            ('sep_conv_3x3', 3), ('dil_conv_3x3', 1), ('max_pool_3x3', 0), ('sep_conv_3x3', 4)],
    reduce_concat=[2, 3, 4, 5])



#!/usr/bin/env python3

from .badnet import BadNet

from autovul.vision.environ import env
from autovul.base.utils import to_tensor, tanh_func
from autovul.base.utils import AverageMeter

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import TensorDataset
import argparse

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import torch.utils.data

mse_criterion = nn.MSELoss()


class LatentBackdoor(BadNet):
    r"""
    Latent Backdoor Attack is described in detail in the paper `Latent Backdoor`_ by Yuanshun Yao.
    Similar to TrojanNN, Latent Backdoor Attack proposes a method to preprocess the trigger pattern.
    The loss function invloves the distance in feature space.

    .. _Latent Backdoor:
        http://people.cs.uchicago.edu/~huiyingli/publication/fr292-yaoA.pdf
    """
    name: str = 'latent_backdoor'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--class_sample_num', type=int,
                           help='the number of sampled images per class, defaults to config[latent_backdoor][class_sample_num][dataset]=100')
        group.add_argument('--mse_weight', type=float,
                           help='the weight of mse loss during retraining, defaults to config[latent_backdoor][mse_weight][dataset]=100')
        group.add_argument('--preprocess_layer',
                           help='the chosen feature layer patched by trigger, defaults to "features"')
        group.add_argument('--preprocess_epoch', type=int, help='preprocess optimization epoch')
        group.add_argument('--preprocess_lr', type=float, help='preprocess learning rate')
        return group

    def __init__(self, class_sample_num: int = 100, mse_weight=0.5,
                 preprocess_layer: str = 'flatten', preprocess_epoch: int = 100, preprocess_lr: float = 0.1,
                 **kwargs):
        super().__init__(**kwargs)

        self.param_list['latent_backdoor'] = ['class_sample_num', 'mse_weight',
                                              'preprocess_layer', 'preprocess_epoch', 'preprocess_lr']
        self.class_sample_num: int = class_sample_num
        self.mse_weight: float = mse_weight

        self.preprocess_layer: str = preprocess_layer
        self.preprocess_epoch: int = preprocess_epoch
        self.preprocess_lr: float = preprocess_lr

        self.avg_target_feats: torch.Tensor = None

    def attack(self, **kwargs):
        print('Sample Data')
        data = self.sample_data()
        print('Calculate Average Target Features')
        self.avg_target_feats = self.get_avg_target_feats(data)
        print('Preprocess Mark')
        self.preprocess_mark(data=data)
        print('Retrain')
        return super().attack(**kwargs)

    def sample_data(self) -> dict[str, tuple[torch.Tensor, torch.Tensor]]:
        other_classes = list(range(self.dataset.num_classes))
        other_classes.pop(self.target_class)
        other_x, other_y = [], []
        for _class in other_classes:
            loader = self.dataset.get_dataloader(mode='train', batch_size=self.class_sample_num, class_list=[_class],
                                                 shuffle=True, num_workers=0, pin_memory=False)
            _input, _label = next(iter(loader))
            other_x.append(_input)
            other_y.append(_label)
        other_x = torch.cat(other_x)
        other_y = torch.cat(other_y)
        target_loader = self.dataset.get_dataloader(mode='train', batch_size=self.class_sample_num, class_list=[self.target_class],
                                                    shuffle=True, num_workers=0, pin_memory=False)
        target_x, target_y = next(iter(target_loader))
        data = {
            'other': (other_x, other_y),
            'target': (target_x, target_y)
        }
        return data

    def get_avg_target_feats(self, data_dict: dict[str, tuple[torch.Tensor, torch.Tensor]]):
        with torch.no_grad():
            if self.dataset.data_shape[1] > 100:
                target_x, target_y = data_dict['target']
                dataset = TensorDataset(target_x, target_y)
                loader = torch.utils.data.DataLoader(dataset=dataset, batch_size=self.dataset.batch_size // max(env['num_gpus'], 1),
                                                     shuffle=True, num_workers=0, pin_memory=False)
                feat_list = []
                for data in loader:
                    target_x, _ = self.model.get_data(data)
                    feat_list.append(self.model.get_layer(target_x, layer_output=self.preprocess_layer).detach().cpu())
                avg_target_feats = torch.cat(feat_list).mean(dim=0)
                avg_target_feats = avg_target_feats.to(target_x.device)
            else:
                target_x, _ = self.model.get_data(data_dict['target'])
                avg_target_feats = self.model.get_layer(target_x, layer_output=self.preprocess_layer).mean(dim=0)
        return avg_target_feats.detach()

    def preprocess_mark(self, data: dict[str, tuple[torch.Tensor, torch.Tensor]]):
        other_x, _ = data['other']
        other_set = TensorDataset(other_x)
        other_loader = self.dataset.get_dataloader(mode='train', dataset=other_set, num_workers=0)

        atanh_mark = torch.randn_like(self.mark.mark) * self.mark.mask
        atanh_mark.requires_grad_()
        self.mark.mark = tanh_func(atanh_mark)
        optimizer = optim.Adam([atanh_mark], lr=self.preprocess_lr)
        optimizer.zero_grad()

        losses = AverageMeter('Loss', ':.4e')
        for _epoch in range(self.preprocess_epoch):
            loader = other_loader
            for (batch_x, ) in loader:
                poison_x = self.mark.add_mark(to_tensor(batch_x))
                loss = self.loss_mse(poison_x)
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                self.mark.mark = tanh_func(atanh_mark)
                losses.update(loss.item(), n=len(batch_x))
        atanh_mark.requires_grad = False
        self.mark.mark.detach_()

    # -------------------------------- Loss Utils ------------------------------ #
    def loss_fn(self, _input: torch.Tensor, _label: torch.Tensor, **kwargs) -> torch.Tensor:
        loss_ce = self.model.loss(_input, _label, **kwargs)
        poison_input = self.mark.add_mark(_input)
        loss_mse = self.loss_mse(poison_input)
        return loss_ce + self.mse_weight * loss_mse

    def loss_mse(self, poison_x: torch.Tensor) -> torch.Tensor:
        other_feats = self.model.get_layer(poison_x, layer_output=self.preprocess_layer)
        return mse_criterion(other_feats, self.avg_target_feats)



#!/usr/bin/env python3

from autovul.vision.datasets.imageset import ImageSet

from .cifar import CIFAR10, CIFAR100
from .downsampled_imagenet import ImageNet16, ImageNet32
from .mnist import MNIST

__all__ = ['MNIST', 'CIFAR10', 'CIFAR100', 'ImageNet16', 'ImageNet32']

class_dict: dict[str, ImageSet] = {
    'mnist': MNIST,
    'cifar10': CIFAR10,
    'cifar100': CIFAR100,
    'imagenet16': ImageNet16,
    'imagenet32': ImageNet32,
}



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch.nn as nn
import torchvision.models
from torchvision.models.vgg import model_urls as urls


class _VGG(_ImageModel):

    def __init__(self, name: str = 'vgg', **kwargs):
        if '_comp' in name:
            comp_dict = {'conv_dim': 512, 'fc_depth': 3, 'fc_dim': 512}
            if '_s' in name:
                comp_dict['fc_depth'] = 1
                comp_dict['fc_dim'] = 0
            for key, value in comp_dict.items():
                if key not in kwargs.keys():
                    kwargs[key] = value
        super().__init__(**kwargs)
        class_name = name.replace('_comp', '').replace('_s', '')
        ModelClass: type[torchvision.models.VGG] = getattr(torchvision.models, class_name)
        _model = ModelClass(num_classes=self.num_classes)
        self.features: nn.Sequential = _model.features
        if '_comp' in name:
            self.pool = nn.AdaptiveAvgPool2d((1, 1))
        else:
            self.pool = _model.avgpool   # nn.AdaptiveAvgPool2d((7, 7))
            self.classifier = _model.classifier
        # nn.Sequential(
        #     nn.Linear(512 * 7 * 7, 4096),
        #     nn.ReLU(True),
        #     nn.Dropout(),
        #     nn.Linear(4096, 4096),
        #     nn.ReLU(True),
        #     nn.Dropout(),
        #     nn.Linear(4096, num_classes),
        # ))


class VGG(ImageModel):
    available_models = ['vgg', 'vgg_bn', 'vgg_comp', 'vgg_bn_comp',
                        'vgg11', 'vgg13', 'vgg16', 'vgg19',
                        'vgg11_bn', 'vgg13_bn', 'vgg16_bn', 'vgg19_bn',
                        'vgg11_comp', 'vgg13_comp', 'vgg16_comp', 'vgg19_comp',
                        'vgg11_bn_comp', 'vgg13_bn_comp', 'vgg16_bn_comp', 'vgg19_bn_comp'
                        'vgg11_comp_s', 'vgg13_comp_s', 'vgg16_comp_s', 'vgg19_comp_s',
                        'vgg11_bn_comp_s', 'vgg13_bn_comp_s', 'vgg16_bn_comp_s', 'vgg19_bn_comp_s']

    model_urls = urls

    def __init__(self, name: str = 'vgg', layer: int = 13,
                 model: type[_VGG] = _VGG, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)



#!/usr/bin/env python3

# https://github.com/facebookresearch/LaMCTS

from .darts.genotypes import Genotype

operations = ['sep_conv_3x3',
              'max_pool_3x3',
              'skip_connect',
              'sep_conv_5x5']


def gen_code_from_list(sample: list[int], node_num: int = None) -> list[list[int]]:
    node_num = node_num if node_num is not None else len(sample) // 4
    return [[sample[i * 2 + j + (0 if j <= 1 else (node_num - 1) * 2)]
             for j in range(4)] for i in range(node_num)]


def translator(code: list[list[int]], max_node: int = None) -> Genotype:
    # input: code type
    # output: geno type
    max_node = max_node if max_node is not None else len(code)
    normal: list[tuple[str, int]] = []
    reduce: list[tuple[str, int]] = []
    normal_concat: list[int] = list(range(max_node + 2))
    reduce_concat: list[int] = list(range(max_node + 2))
    for cell, concat in zip([normal, reduce], [normal_concat, reduce_concat]):
        for block in range(len(code)):
            cell.append((operations[code[block][0]], code[block][2]))
            cell.append((operations[code[block][1]], code[block][3]))
            if code[block][2] in concat:
                concat.remove(code[block][2])
            if code[block][3] in concat:
                concat.remove(code[block][3])
    if 0 in reduce_concat:
        reduce_concat.remove(0)
    if 1 in reduce_concat:
        reduce_concat.remove(1)
    return Genotype(normal=normal, normal_concat=normal_concat, reduce=reduce, reduce_concat=reduce_concat)



#!/usr/bin/env python3

# https://github.com/rwightman/pytorch-image-models/tree/master/timm/models
# https://github.com/kuangliu/pytorch-cifar/blob/master/models/dla.py

'''DLA in PyTorch.

Reference:
    Deep Layer Aggregation. https://arxiv.org/abs/1707.06484
'''

import torch
import torch.nn as nn
from collections import OrderedDict
from torchvision.models.resnet import conv1x1, conv3x3

from collections.abc import Callable
from typing import Optional


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(self, inplanes: int, planes: int, stride: int = 1, dilation: int = 1) -> None:
        super().__init__()
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        self.conv1 = conv3x3(inplanes, planes, stride=stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.stride = stride

    def forward(self, x: torch.Tensor, residual: torch.Tensor = None) -> torch.Tensor:
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion: int = 2

    def __init__(self, inplanes: int, planes: int, stride: int = 1,
                 groups: int = 1, base_width: int = 64, dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * groups * (base_width / 64.) / self.expansion)

        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes)
        self.bn3 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.stride = stride

    def forward(self, x: torch.Tensor, residual: torch.Tensor = None) -> torch.Tensor:
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += residual
        out = self.relu(out)

        return out


class Root(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 1, residual: bool = False) -> None:
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size,
                              stride=1, padding=(kernel_size - 1) // 2,
                              bias=False)
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.residual = residual

    def forward(self, *xs: list[torch.Tensor]) -> torch.Tensor:
        identity = xs[0]

        out = torch.cat(xs, dim=1)
        out = self.conv(out)
        out = self.bn(out)
        out = self.relu(out)

        if self.residual:
            out += identity

        return out


class Tree(nn.Module):
    def __init__(self, block: type[nn.Module], in_channels: int, out_channels: int,
                 levels: int, stride: int = 1,
                 level_root=False, root_dim: int = 0, root_kernel_size: int = 1,
                 dilation: int = 1, root_residual: bool = False, **kwargs):
        super().__init__()
        if root_dim == 0:
            root_dim = 2 * out_channels
        if level_root:
            root_dim += in_channels
        self.downsample: nn.Module = None
        self.project: nn.Module = None
        if stride > 1:
            self.downsample = nn.MaxPool2d(stride, stride=stride)

        if levels == 1:
            self.tree1 = block(in_channels, out_channels, stride,
                               dilation=dilation, **kwargs)
            self.tree2 = block(out_channels, out_channels, 1,
                               dilation=dilation, **kwargs)
            self.root = Root(root_dim, out_channels, root_kernel_size,
                             root_residual)
            if in_channels != out_channels:
                # NOTE the official impl/weights have project layers in levels > 1 case that are never
                # used, I've moved the project layer here to avoid wasted params but old checkpoints will
                # need strict=False while loading.
                self.project = nn.Sequential(
                    nn.Conv2d(in_channels, out_channels,
                              kernel_size=1, stride=1, bias=False),
                    nn.BatchNorm2d(out_channels)
                )
        else:
            self.tree1 = Tree(block, in_channels, out_channels, levels - 1,
                              stride, root_dim=0,
                              root_kernel_size=root_kernel_size,
                              dilation=dilation, root_residual=root_residual, **kwargs)
            self.tree2 = Tree(block, out_channels, out_channels, levels - 1,
                              root_dim=root_dim + out_channels,
                              root_kernel_size=root_kernel_size,
                              dilation=dilation, root_residual=root_residual, **kwargs)
        self.level_root = level_root
        self.root_dim = root_dim
        self.levels = levels

    def forward(self, x: torch.Tensor, residual: torch.Tensor = None,
                children: list[torch.Tensor] = None) -> torch.Tensor:
        children = [] if children is None else children
        bottom = self.downsample(x) if self.downsample else x
        residual = self.project(bottom) if self.project else bottom
        if self.level_root:
            children.append(bottom)
        x1 = self.tree1(x, residual)
        if self.levels == 1:
            x2 = self.tree2(x1)
            x = self.root(x2, x1, *children)
        else:
            children.append(x1)
            x = self.tree2(x1, children=children)
        return x


class DLA(nn.Module):
    def __init__(self, block: nn.Module, levels: list[int], channels: list[int], num_classes: int = 1000,
                 root_residual: bool = False, strides: list[int] = [2, 2, 2, 2], **kwargs):
        super().__init__()
        self.features = nn.Sequential(OrderedDict([
            ('stem', nn.Sequential(OrderedDict([
                ('conv1', nn.Conv2d(3, channels[0], kernel_size=7, stride=1, padding=3, bias=False)),    # stride=1
                ('bn1', nn.BatchNorm2d(16)),
                ('relu', nn.ReLU(True)),
                # ('maxpool', nn.MaxPool2d(kernel_size=3, stride=2, padding=1)),
            ]))),
            ('layer1', nn.Sequential(OrderedDict([
                ('conv1', conv3x3(channels[0], channels[0])),
                ('bn1', nn.BatchNorm2d(16)),
                ('relu', nn.ReLU(True))
            ]))),
            ('layer2', nn.Sequential(OrderedDict([
                ('conv1', conv3x3(channels[0], channels[1], stride=2)),
                ('bn1', nn.BatchNorm2d(32)),
                ('relu', nn.ReLU(True))
            ]))),
            ('layer3', Tree(block, channels[1], channels[2], levels=levels[0], stride=strides[0],
                            level_root=False, root_residual=root_residual, **kwargs)),
            ('layer4', Tree(block, channels[2], channels[3], levels=levels[1], stride=strides[1],
                            level_root=True, root_residual=root_residual, **kwargs)),
            ('layer5', Tree(block, channels[3], channels[4], levels=levels[2], stride=strides[2],
                            level_root=True, root_residual=root_residual, **kwargs)),
            ('layer6', Tree(block, channels[4], channels[5], levels=levels[3], stride=strides[3],
                            level_root=True, root_residual=root_residual, **kwargs)),
        ]))
        self.pool = nn.AdaptiveAvgPool2d(1)
        self.flatten = nn.Flatten()
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', nn.Linear(channels[-1], num_classes))
        ]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.features(x)
        out = self.pool(out)
        out = self.flatten(out)
        out = self.classifier(out)
        return out


def dla34(**kwargs) -> DLA:
    return DLA(BasicBlock, levels=[1, 2, 2, 1], channels=[16, 32, 64, 128, 256, 512], **kwargs)


def dla46_c(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 2, 2, 1], channels=[16, 32, 64, 64, 128, 256], **kwargs)


def dla46x_c(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 2, 2, 1], channels=[16, 32, 64, 64, 128, 256],
               groups=32, base_width=4, **kwargs)


def dla60x_c(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 2, 3, 1], channels=[16, 32, 64, 64, 128, 256],
               groups=32, base_width=4, **kwargs)


def dla60(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 2, 3, 1], channels=[16, 32, 128, 256, 512, 1024], **kwargs)


def dla60x(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 2, 3, 1], channels=[16, 32, 128, 256, 512, 1024],
               groups=32, base_width=4, **kwargs)


def dla102(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 3, 4, 1], channels=[16, 32, 128, 256, 512, 1024],
               root_residual=True, **kwargs)


def dla102x(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 3, 4, 1], channels=[16, 32, 128, 256, 512, 1024],
               groups=32, base_width=4, root_residual=True, **kwargs)


def dla102x2(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[1, 3, 4, 1], channels=[16, 32, 128, 256, 512, 1024],
               groups=64, base_width=4, root_residual=True, **kwargs)


def dla169(**kwargs) -> DLA:
    return DLA(Bottleneck, levels=[2, 3, 5, 1], channels=[16, 32, 128, 256, 512, 1024],
               root_residual=True, **kwargs)



#!/usr/bin/env python3

from .darts import _DARTS, DARTS
from autovul.vision.datasets import ImageSet
from autovul.vision.utils.model_archs.lanet import gen_code_from_list, translator

import torch
import torch.hub
from torchvision.datasets.utils import download_file_from_google_drive

import os
import zipfile
from collections import OrderedDict


cifar_arch = [2, 2, 0, 2, 1, 2, 0, 2, 2, 3, 2, 1, 2, 0, 0, 1, 1, 1, 2, 1, 1, 0, 3, 4, 3, 0, 3, 1]


class LaNet(DARTS):
    available_models = ['lanet']
    model_urls = {'cifar10': '1bZsEoG-sroVyYR4F_2ozGLA5W50CT84P', }

    def __init__(self, name: str = 'lanet', layers: int = 24, C: int = 128,
                 arch: list[int] = cifar_arch, model: type[_DARTS] = _DARTS, **kwargs):
        genotype = translator(gen_code_from_list(arch))
        self.arch = arch
        super().__init__(name=name, layers=layers, C=C, genotype=genotype, model=model, std_conv=True, **kwargs)
        self.param_list['lanet'] = ['arch']

    def get_official_weights(self, dataset: str = None, **kwargs) -> OrderedDict[str, torch.Tensor]:
        if dataset is None and isinstance(self.dataset, ImageSet):
            dataset = self.dataset.name
        folder_path = os.path.join(torch.hub.get_dir(), 'lanet')
        file_path = os.path.join(folder_path, f'lanet_{dataset}.pt')
        if not os.path.exists(file_path):
            zip_file_name = 'temp.zip'
            zip_path = os.path.join(folder_path, zip_file_name)
            download_file_from_google_drive(file_id=self.model_urls[dataset], root=folder_path, filename=zip_file_name)
            with zipfile.ZipFile(zip_path, 'r') as zf:
                data = zf.read('lanas_128_99.03/top1.pt')
            with open(file_path, 'wb') as f:
                f.write(data)
            os.remove(zip_path)
        print('get official model weights from Google Drive: ', self.model_urls[dataset])
        _dict: OrderedDict[str, torch.Tensor] = torch.load(file_path, map_location='cpu')
        if 'model_state_dict' in _dict.keys():
            _dict = _dict['model_state_dict']
        new_dict: OrderedDict[str, torch.Tensor] = self.state_dict()
        old_keys = list(_dict.keys())
        new_keys = list(new_dict.keys())
        new2old: dict[str, str] = {}
        i = 0
        j = 0
        while(i < len(new_keys) and j < len(old_keys)):
            if 'auxiliary_head' not in new_keys[i] and 'auxiliary_head' in old_keys[j]:
                j += 1
                continue
            new2old[new_keys[i]] = old_keys[j]
            i += 1
            j += 1
        for i, key in enumerate(new_keys):
            new_dict[key] = _dict[new2old[key]]
        return new_dict



#!/usr/bin/env python3

from autovul.base.utils.output import ansi, prints
from autovul.base.utils import repeat_to_batch

import torch
import torch.nn as nn
import torchvision.transforms as transforms

from typing import Iterator

__all__ = ['get_all_layer', 'get_layer', 'get_layer_name', 'summary', 'activate_params', 'accuracy', 'generate_target']

filter_tuple: tuple[nn.Module] = (transforms.Normalize,
                                  nn.Dropout, nn.BatchNorm2d,
                                  nn.ReLU, nn.Sigmoid)


def get_layer_name(module: nn.Module, depth: int = -1, prefix: str = '',
                   use_filter: bool = True, repeat: bool = False,
                   seq_only: bool = False, init: bool = True) -> list[str]:
    layer_name_list: list[str] = []
    if init or (not seq_only or isinstance(module, nn.Sequential)) and depth != 0:
        for name, child in module.named_children():
            full_name = prefix + ('.' if prefix else '') + name  # prefix=full_name
            layer_name_list.extend(get_layer_name(child, depth - 1, full_name,
                                                  use_filter, repeat, seq_only, init=False))
    if prefix and (not use_filter or filter_layer(module)) \
            and (repeat or depth == 0 or not isinstance(module, nn.Sequential)):
        layer_name_list.append(prefix)
    return layer_name_list


def get_all_layer(module: nn.Module, x: torch.Tensor,
                  layer_input: str = 'input', depth: int = 0,
                  prefix='', use_filter: bool = True, repeat: bool = False,
                  seq_only: bool = True, verbose: int = 0) -> dict[str, torch.Tensor]:
    layer_name_list = get_layer_name(module, depth=depth, prefix=prefix, use_filter=False)
    if layer_input == 'input':
        layer_input = 'record'
    elif layer_input not in layer_name_list:
        print('Model Layer Name List: ', layer_name_list)
        print('Input layer: ', layer_input)
        raise ValueError('Layer name not in model')
    if verbose:
        print(f'{ansi["green"]}{"layer name":<50s}{"output shape":<20}{"module information"}{ansi["reset"]}')
    return _get_all_layer(module, x, layer_input, depth,
                          prefix, use_filter, repeat,
                          seq_only, verbose=verbose, init=True)[0]


def _get_all_layer(module: nn.Module, x: torch.Tensor, layer_input: str = 'record', depth: int = 0,
                   prefix: str = '', use_filter: bool = True, repeat: bool = False,
                   seq_only: bool = True, verbose: int = 0, init: bool = False
                   ) -> tuple[dict[str, torch.Tensor], torch.Tensor]:
    _dict: dict[str, torch.Tensor] = {}
    if init or (not seq_only or isinstance(module, nn.Sequential)) and depth != 0:
        for name, child in module.named_children():
            full_name = prefix + ('.' if prefix else '') + name  # prefix=full_name
            if layer_input == 'record' or layer_input.startswith(f'{full_name}.'):
                sub_dict, x = _get_all_layer(child, x, layer_input, depth - 1,
                                             full_name, use_filter, repeat, seq_only, verbose)
                _dict.update(sub_dict)
                layer_input = 'record'
            elif layer_input == full_name:
                layer_input = 'record'
    else:
        x = module(x)
    if prefix and (not use_filter or filter_layer(module)) \
            and (repeat or depth == 0 or not isinstance(module, nn.Sequential)):
        _dict[prefix] = x.clone()
        if verbose:
            shape_str = str(list(x.shape))
            module_str = ''
            if verbose == 1:
                module_str = module.__class__.__name__
            elif verbose == 2:
                module_str = type(module)
            elif verbose == 3:
                module_str = str(module).split('\n')[0].removesuffix('(')
            else:
                module_str = str(module)
            print(f'{ansi["blue_light"]}{prefix:<50s}{ansi["reset"]}{ansi["yellow"]}{shape_str:<20}{ansi["reset"]}{module_str}')
    return _dict, x


def get_layer(module: nn.Module, x: torch.Tensor, layer_output: str = 'output',
              layer_input: str = 'input', prefix: str = '',
              layer_name_list: list[str] = None,
              seq_only: bool = True) -> torch.Tensor:
    if layer_input == 'input' and layer_output == 'output':
        return module(x)
    if layer_name_list is None:
        layer_name_list = get_layer_name(module, use_filter=False, repeat=True)
        layer_name_list.insert(0, 'input')
        layer_name_list.append('output')
    if layer_input not in layer_name_list or layer_output not in layer_name_list \
            or layer_name_list.index(layer_input) > layer_name_list.index(layer_output):
        print('Model Layer Name List: \n', layer_name_list)
        print('Input  layer: ', layer_input)
        print('Output layer: ', layer_output)
        raise ValueError('Layer name not correct')
    if layer_input == 'input':
        layer_input = 'record'
    return _get_layer(module, x, layer_output, layer_input, prefix, seq_only, init=True)


def _get_layer(module: nn.Module, x: torch.Tensor, layer_output: str = 'output',
               layer_input: str = 'record', prefix: str = '',
               seq_only: bool = True, init: bool = False) -> torch.Tensor:
    if init or (not seq_only or isinstance(module, nn.Sequential)):
        for name, child in module.named_children():
            full_name = prefix + ('.' if prefix else '') + name  # prefix=full_name
            if layer_input == 'record' or layer_input.startswith(f'{full_name}.'):
                x = _get_layer(child, x, layer_output, layer_input, full_name, seq_only)
                layer_input = 'record'
            elif layer_input == full_name:
                layer_input = 'record'
            if layer_output.startswith(full_name):
                return x
    else:
        x = module(x)
    return x


def filter_layer(module: nn.Module, filter_tuple: tuple[nn.Module] = filter_tuple) -> bool:
    if isinstance(module, filter_tuple):
        return False
    return True


def summary(module: nn.Module, depth: int = 0, verbose: bool = True,
            indent: int = 0, tree_length: int = None, indent_atom: int = 12) -> None:
    tree_length = tree_length if tree_length is not None else indent_atom * (depth + 1)
    if depth > 0:
        for name, child in module.named_children():
            _str = f'{ansi["blue_light"]}{name}{ansi["reset"]}'
            if verbose:
                _str = _str.ljust(tree_length - indent +
                                  len(ansi['blue_light']) + len(ansi['reset']))
                _str += str(child).split('\n')[0].removesuffix('(')
            prints(_str, indent=indent)
            summary(child, depth=depth - 1, indent=indent + indent_atom,
                    verbose=verbose, tree_length=tree_length)


def activate_params(module: nn.Module, params: Iterator[nn.Parameter]) -> None:
    for param in module.parameters():
        param.requires_grad_(False)
    for param in params:
        param.requires_grad_()


def accuracy(_output: torch.Tensor, _label: torch.Tensor, num_classes: int,
             topk: tuple[int] = (1, 5)) -> list[float]:
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = min(max(topk), num_classes)
        batch_size = _label.size(0)
        _, pred = _output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(_label[None])
        res: list[float] = []
        for k in topk:
            if k > num_classes:
                res.append(100.0)
            else:
                correct_k = float(correct[:k].sum(dtype=torch.float32))
                res.append(correct_k * (100.0 / batch_size))
        return res


def generate_target(module: nn.Module, _input: torch.Tensor, idx: int = 1, same: bool = False) -> torch.Tensor:
    with torch.no_grad():
        _output: torch.Tensor = module(_input)
    target = _output.argsort(dim=-1, descending=True)[:, idx]
    if same:
        target = repeat_to_batch(target.mode(dim=0)[0], len(_input))
    return target



#!/usr/bin/env python3

# https://github.com/facebookresearch/GradientEpisodicMemory/blob/master/model/common.py

# Copyright 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn
from torchvision.models.resnet import BasicBlock, conv3x3


class ResNetS(nn.Module):
    def __init__(self, block: BasicBlock, num_blocks: list[int], num_classes: int, nf: int):
        super().__init__()
        self.in_planes = nf

        self.conv1 = conv3x3(3, nf * 1)
        self.bn1 = nn.BatchNorm2d(nf * 1)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, nf * 1, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, nf * 2, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, nf * 4, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, nf * 8, num_blocks[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(nf * 8 * block.expansion, num_classes)

    def _make_layer(self, block: BasicBlock, planes: int, num_blocks: list[int], stride: int):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x


def resnet_s(nclasses: int, nf: int = 20):
    return ResNetS(BasicBlock, [2, 2, 2, 2], nclasses, nf)



#!/usr/bin/env python3

from autovul.vision.optim import PGDoptimizer
from autovul.base.attacks import Attack
from autovul.base.utils import to_list
from autovul.base.utils.output import prints, ansi
from autovul.base.utils.logger import SmoothedValue

import torch
import argparse
from collections.abc import Callable
from typing import Union
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from autovul.vision.datasets import ImageSet
    from autovul.vision.models import ImageModel


def get_func_key(func: Callable[..., torch.Tensor]) -> str:
    keys = func.__code__.co_varnames
    valid_keys = [key for key in keys if 'target' in key or 'label' in key or 'y' in key or 'Y' in key]
    assert len(valid_keys) == 1, keys
    return valid_keys[0]


class PGD(Attack, PGDoptimizer):
    r"""PGD Adversarial Attack.
    Args:
        pgd_alpha (float): learning rate :math:`\alpha`. Default: :math:`\frac{3}{255}`.
        pgd_eps (float): the perturbation threshold :math:`\epsilon` in input space. Default: :math:`\frac{8}{255}`.
    """

    name: str = 'pgd'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--pgd_alpha', type=float, help='PGD learning rate per step, defaults to 2.0/255')
        group.add_argument('--pgd_eps', type=float, help='Projection norm constraint, defaults to 8.0/255')
        group.add_argument('--iteration', type=int, help='Attack Iteration, defaults to 7')
        group.add_argument('--stop_threshold', type=float, help='early stop confidence, defaults to 0.99')
        group.add_argument('--target_class', type=int, help='Do not set it if using target_idx')
        group.add_argument('--target_idx', type=int, help='Target label order in original classification, defaults to -1 '
                           '(0 for untargeted attack, 1 for most possible class, -1 for most unpossible class)')
        group.add_argument('--test_num', type=int, help='total number of test examples for PGD, defaults to 1000.')
        group.add_argument('--num_restart', type=int,
                           help='number of random init for PGD, defaults to 0 (without random initialization).')
        group.add_argument('--require_class', action='store_true',
                           help='Require class to be desired in early_stop_check.')

        group.add_argument('--grad_method', choices=['white', 'nes', 'sgd', 'hess', 'zoo'],
                           help='gradient estimation method, defaults to "white"')
        group.add_argument('--query_num', type=int,
                           help='query numbers for black box gradient estimation, defaults to 100.')
        group.add_argument('--sigma', type=float,
                           help='gaussian sampling std for black box gradient estimation, defaults to 1e-3')
        return group

    def __init__(self, target_class: int = None, target_idx: int = -1, test_num: int = 1000, num_restart: int = 0,
                 require_class: bool = False, **kwargs):
        self.target_class = target_class
        self.target_idx = target_idx
        self.test_num = test_num
        self.num_restart = num_restart
        self.require_class = require_class
        kwargs.update(random_init=bool(num_restart))
        super().__init__(**kwargs)
        self.param_list['pgd_attack'] = ['target_class', 'target_idx', 'test_num', 'num_restart', 'require_class']
        self.dataset: ImageSet
        self.model: 'ImageModel'

    def attack(self, verbose: int = 1, **kwargs) -> tuple[float, float]:
        loader = self.dataset.get_dataloader(mode='test', batch_size=self.dataset.valid_batch_size,
                                             shuffle=True)
        fmt_str = '{global_avg:7.3f} ({min:7.3f}  {max:7.3f})'
        total_adv_target_conf = SmoothedValue(fmt=fmt_str)
        total_org_target_conf = SmoothedValue(fmt=fmt_str)
        succ_adv_target_conf = SmoothedValue(fmt=fmt_str)

        total_adv_org_conf = SmoothedValue(fmt=fmt_str)
        total_org_org_conf = SmoothedValue(fmt=fmt_str)
        succ_adv_org_conf = SmoothedValue(fmt=fmt_str)

        total_iter_list = SmoothedValue(fmt=fmt_str)
        succ_iter_list = SmoothedValue(fmt=fmt_str)

        succ_idx_list: list[int] = []
        for data in loader:
            rest_length = self.test_num - total_adv_target_conf.count
            if rest_length <= 0:
                break
            _input, _label = self.model.remove_misclassify(data)
            if len(_label) == 0:
                continue

            if len(_label) > rest_length:
                _input = _input[:rest_length]
                _label = _label[:rest_length]
            target = self.generate_target(_input, idx=self.target_idx) if self.target_class is None \
                else self.target_class * torch.ones_like(_label)
            adv_input = _input.clone().detach()
            iter_list = -torch.ones(len(_label), dtype=torch.long)
            current_idx = torch.arange(len(iter_list))
            for _ in range(max(self.num_restart, 1)):
                temp_adv_input, temp_iter_list = self.optimize(_input[current_idx],
                                                               target=target[current_idx], **kwargs)
                adv_input[current_idx] = temp_adv_input
                iter_list[current_idx] = temp_iter_list
                fail_idx = iter_list == -1
                if (~fail_idx).all():
                    break
                current_idx = current_idx[fail_idx]
            for i, _iter in enumerate(iter_list):
                if _iter != -1:
                    succ_idx_list.append(total_iter_list.count + i)
            adv_target_conf = self.model.get_target_prob(adv_input, target)
            adv_org_conf = self.model.get_target_prob(adv_input, _label)
            org_target_conf = self.model.get_target_prob(_input, target)
            org_org_conf = self.model.get_target_prob(_input, _label)

            total_adv_target_conf.update_list(to_list(adv_target_conf))
            total_adv_org_conf.update_list(to_list(adv_org_conf))
            succ_adv_target_conf.update_list(to_list(adv_target_conf[iter_list != -1]))
            succ_adv_org_conf.update_list(to_list(adv_org_conf[iter_list != -1]))
            total_org_target_conf.update_list(to_list(org_target_conf))
            total_org_org_conf.update_list(to_list(org_org_conf))

            total_iter_list.update_list(to_list(torch.where(iter_list != -1, iter_list, 2 * self.iteration)))
            succ_iter_list.update_list(to_list(iter_list[iter_list != -1]))
            if verbose >= 3:
                prints(f'{ansi["green"]}{succ_iter_list.count} / {total_iter_list.count}{ansi["reset"]}')
            if verbose >= 4:
                prints(f'{total_iter_list=:}', indent=4)
                prints(f'{succ_iter_list=:}', indent=4)
                prints()
                prints('-------------------------------------------------', indent=4)
                prints(f'{ansi["yellow"]}Target Class:{ansi["reset"]}', indent=4)
                prints(f'{total_adv_target_conf=:}', indent=8)
                prints(f'{total_org_target_conf=:}', indent=8)
                prints(f'{succ_adv_target_conf=:}', indent=8)
                prints()
                prints('-------------------------------------------------', indent=4)
                prints(f'{ansi["yellow"]}Original Class:{ansi["reset"]}', indent=4)
                prints(f'{total_adv_org_conf=:}', indent=8)
                prints(f'{total_org_org_conf=:}', indent=8)
                prints(f'{succ_adv_org_conf=:}', indent=8)
        if verbose:
            prints(f'{ansi["green"]}{succ_iter_list.count} / {total_iter_list.count}{ansi["reset"]}')
            prints(succ_idx_list)
        if verbose >= 2:
            prints(f'{total_iter_list=:}', indent=4)
            prints(f'{succ_iter_list=:}', indent=4)
            prints()
            prints('-------------------------------------------------', indent=4)
            prints(f'{ansi["yellow"]}Target Class:{ansi["reset"]}', indent=4)
            prints(f'{total_adv_target_conf=:}', indent=8)
            prints(f'{total_org_target_conf=:}', indent=8)
            prints(f'{succ_adv_target_conf=:}', indent=8)
            prints()
            prints('-------------------------------------------------', indent=4)
            prints(f'{ansi["yellow"]}Original Class:{ansi["reset"]}', indent=4)
            prints(f'{total_adv_org_conf=:}', indent=8)
            prints(f'{total_org_org_conf=:}', indent=8)
            prints(f'{succ_adv_org_conf=:}', indent=8)
        return float(succ_iter_list.count) / total_iter_list.count, total_iter_list.global_avg

    def optimize(self, _input: torch.Tensor, target: Union[torch.Tensor, int] = None, target_idx: int = None,
                 loss_fn: Callable[..., torch.Tensor] = None,
                 require_class: bool = None,
                 loss_kwargs: dict[str, torch.Tensor] = {},
                 *args, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        if len(_input) == 0:
            return _input, None
        target_idx = self.target_idx if target_idx is None else target_idx
        if target is None:
            target = self.generate_target(_input, idx=self.target_idx) if self.target_class is None \
                else self.target_class * torch.ones(len(_input), dtype=torch.long, device=_input.device)
        elif isinstance(target, int):
            target = target * torch.ones(len(_input), dtype=torch.long, device=_input.device)
        else:
            assert isinstance(target, torch.Tensor)
        if loss_fn is None and self.loss_fn is None:
            untarget_condition = self.target_class is None and self.target_idx == 0

            def _loss_fn(_input: torch.Tensor, target: torch.Tensor, reduction: str = 'mean', **kwargs):
                loss = self.model.loss(_input, target, reduction=reduction)
                return -loss if untarget_condition else loss
            loss_fn = _loss_fn
        loss_kwargs.update(target=target)
        return super().optimize(_input, target=target,
                                loss_fn=loss_fn, require_class=require_class,
                                loss_kwargs=loss_kwargs,
                                *args, **kwargs)

    def early_stop_check(self, current_idx: torch.Tensor,
                         adv_input: torch.Tensor, target: torch.Tensor,
                         stop_threshold: float = None, require_class: bool = None,
                         *args, **kwargs) -> torch.Tensor:
        stop_threshold = stop_threshold if stop_threshold is not None else self.stop_threshold
        require_class = require_class if require_class is not None else self.require_class
        if self.stop_threshold is None:
            return torch.zeros(len(current_idx), dtype=torch.bool)
        with torch.no_grad():
            _confidence = self.model.get_target_prob(adv_input[current_idx], target[current_idx])
        untarget_condition = self.target_class is None and self.target_idx == 0
        result = _confidence > stop_threshold
        if untarget_condition:
            result = ~result
        if require_class:
            with torch.no_grad():
                _class = self.model.get_class(adv_input[current_idx])
            class_result = _class == target[current_idx]
            if untarget_condition:
                class_result = ~class_result
            result = result.bitwise_and(class_result)
        return result

    def output_info(self, org_input: torch.Tensor, noise: torch.Tensor, target: torch.Tensor,
                    loss_fn: Callable[[torch.Tensor], torch.Tensor] = None, **kwargs):
        super().output_info(org_input=org_input, noise=noise, loss_fn=loss_fn, **kwargs)
        # prints('Original class     : ', to_list(_label), indent=self.indent)
        # prints('Original confidence: ', to_list(_confidence), indent=self.indent)
        with torch.no_grad():
            _confidence = self.model.get_target_prob(org_input + noise, target)
        prints('Target   class     : ', to_list(target), indent=self.indent)
        prints('Target   confidence: ', to_list(_confidence), indent=self.indent)



#!/usr/bin/env python3

from autovul.vision.datasets.imagefolder import ImageFolder

from .imagenet import ImageNet


__all__ = ['ImageNet']

class_dict: dict[str, ImageFolder] = {
    'imagenet': ImageNet,
}



#!/usr/bin/env python3

from autovul.vision.utils.model_archs.darts import Genotype
from copy import deepcopy

org_arch = Genotype(
    normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1),
            ('sep_conv_3x3', 0), ('sep_conv_3x3', 1),
            ('sep_conv_3x3', 1), ('skip_connect', 0),
            ('skip_connect', 0), ('dil_conv_3x3', 2)],
    normal_concat=[2, 3, 4, 5],
    reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1),
            ('skip_connect', 2), ('max_pool_3x3', 1),
            ('max_pool_3x3', 0), ('skip_connect', 2),
            ('skip_connect', 2), ('max_pool_3x3', 1)],
    reduce_concat=[2, 3, 4, 5])


def mitigation_1(arch: Genotype) -> Genotype:
    '''make arch deep'''
    arch = deepcopy(arch)
    for cell in [arch.normal, arch.reduce]:
        for i, (op, idx) in enumerate(cell):
            if i in [0, 1]:
                cell[i] = (op, i)
            else:
                cell[i] = (op, i // 2 + 1)
    return arch


def mitigation_2(arch: Genotype) -> Genotype:
    '''replace skip connect to sep conv'''
    arch = deepcopy(arch)
    for cell in [arch.normal, arch.reduce]:
        for i, (op, idx) in enumerate(cell):
            if op == 'skip_connect':
                cell[i] = ('sep_conv_3x3', idx)
    return arch


def mitigation_3(arch: Genotype) -> Genotype:
    return mitigation_2(mitigation_1(arch))


print('DARTS-i  : ', mitigation_1(org_arch))
print('DARTS-ii : ', mitigation_2(org_arch))
print('DARTS-iii: ', mitigation_3(org_arch))



#!/usr/bin/env python3

import torch
import random
from typing import Union  # TODO: python 3.10


__all__ = ['Cutout']


class Cutout:
    def __init__(self, length: int, fill_values: Union[float, torch.Tensor] = 0.0):
        self.length = length
        self.fill_values = fill_values

    def __call__(self, img: torch.Tensor):
        h, w = img.size(1), img.size(2)
        mask = torch.ones(h, w, dtype=torch.bool, device=img.device)
        y = random.randint(0, h)
        x = random.randint(0, w)
        y1 = max(y - self.length // 2, 0)
        y2 = min(y + self.length // 2, h)
        x1 = max(x - self.length // 2, 0)
        x2 = min(x + self.length // 2, w)
        mask[y1: y2, x1: x2] = False
        return (mask * img + ~mask * self.fill_values).detach()



#!/usr/bin/env python3

from autovul.base.configs import config
from autovul.base.datasets import Dataset
from autovul.base.environ import env
from autovul.base.utils import add_noise, get_name, to_tensor
from autovul.base.utils.model import *
from autovul.base.utils.train import train, validate, compare
from autovul.base.utils.output import ansi, prints
from autovul.base.utils.fim import KFAC, EKFAC

import torch
import torch.nn as nn
from torch.utils import model_zoo
import numpy as np
import os
from collections import OrderedDict
from collections.abc import Iterable    # TODO: callable (many places) (wait for python update)

from typing import TYPE_CHECKING
from typing import Generator, Iterator, Mapping, Optional, Set, Union    # TODO: python 3.10
from autovul.base.configs import Config    # TODO: python 3.10
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import argparse
from collections.abc import Callable
if TYPE_CHECKING:
    import torch.utils.data
# redirect = Indent_Redirect(buffer=True, indent=0)

__all__ = ['Model', 'add_argument', 'create',
           'get_available_models', 'output_available_models', 'get_model_class']


class _Model(nn.Module):
    def __init__(self, num_classes: int = None, **kwargs):
        super().__init__()
        self.define_preprocess(**kwargs)
        self.features = self.define_features(**kwargs)   # feature extractor
        self.pool = nn.AdaptiveAvgPool2d((1, 1))  # average pooling
        self.flatten = nn.Flatten()
        self.classifier = self.define_classifier(num_classes=num_classes, **kwargs)  # classifier
        self.softmax = nn.Softmax(dim=1)

        self.num_classes = num_classes

    def define_preprocess(self, **kwargs):
        pass

    @staticmethod
    def define_features(**kwargs) -> nn.Module:
        return nn.Identity()

    @staticmethod
    def define_classifier(conv_dim: int = 0, num_classes: int = None,
                          fc_depth: int = 0, fc_dim: int = 0,
                          activation: type[nn.Module] = nn.ReLU,
                          dropout: float = 0.5,
                          **kwargs) -> nn.Sequential:
        seq = nn.Sequential()
        if fc_depth <= 0:
            return seq
        dim_list: list[int] = [fc_dim] * (fc_depth - 1)
        dim_list.insert(0, conv_dim)
        activation_name: str = 'none'
        if activation:
            activation_name = activation.__name__.split('.')[-1].lower()
        if fc_depth == 1:
            seq.add_module('fc', nn.Linear(conv_dim, num_classes))
        else:
            for i in range(fc_depth - 1):
                seq.add_module(f'fc{i + 1:d}', nn.Linear(dim_list[i], dim_list[i + 1]))
                if activation:
                    seq.add_module(f'{activation_name}{i + 1:d}', activation(True))
                if dropout > 0:
                    seq.add_module(f'dropout{i + 1:d}', nn.Dropout(p=dropout))
            seq.add_module(f'fc{fc_depth:d}', nn.Linear(fc_dim, num_classes))
        return seq

    # forward method
    # input: (batch_size, channels, height, width)
    # output: (batch_size, logits)
    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        x = self.get_final_fm(x, **kwargs)
        x = self.classifier(x)
        return x

    # input: (batch_size, channels, height, width)
    # output: (batch_size, [feature_map])
    def get_fm(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        return self.features(x)

    def get_final_fm(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        x = self.get_fm(x, **kwargs)
        x = self.pool(x)
        x = self.flatten(x)
        return x


class Model:
    available_models: list[str] = []
    model_urls: dict[str, str] = []

    @staticmethod
    def add_argument(group: argparse._ArgumentGroup):
        group.add_argument('-m', '--model', dest='model_name',
                           help='model name, defaults to config[model][default_model]')
        group.add_argument('--suffix', help='model name suffix, e.g. _adv_train')
        group.add_argument('--pretrain', action='store_true', help='load pretrained weights, defaults to False')
        group.add_argument('--official', action='store_true', help='load official weights, defaults to False')
        group.add_argument('--model_dir', help='directory to contain pretrained models')
        group.add_argument('--randomized_smooth', help='whether to use randomized smoothing, defaults to False')
        group.add_argument('--rs_sigma', type=float, help='randomized smoothing sampling std, defaults to 0.01')
        group.add_argument('--rs_n', type=int, help='randomized smoothing sampling number, defaults to 100')
        return group

    def __init__(self, name: str = 'model', model: Union[type[_Model], _Model] = _Model,
                 dataset: Dataset = None,
                 num_classes: int = None, folder_path: str = None,
                 official: bool = False, pretrain: bool = False,
                 randomized_smooth: bool = False, rs_sigma: float = 0.01, rs_n: int = 100,
                 suffix: str = None, verbose: bool = False, **kwargs):
        self.param_list: dict[str, list[str]] = {}
        self.param_list['model'] = ['folder_path']
        if suffix is not None:
            self.param_list['model'].append('suffix')
        else:
            suffix = ''
        if randomized_smooth:
            self.param_list['model'].extend(['randomized_smooth', 'rs_sigma', 'rs_n'])
        self.name: str = name
        self.dataset = dataset
        self.suffix = suffix
        self.randomized_smooth: bool = randomized_smooth
        self.rs_sigma: float = rs_sigma
        self.rs_n: int = rs_n

        self.folder_path = folder_path
        if folder_path is not None:
            self.folder_path = os.path.normpath(folder_path)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)

        # ------------Auto-------------- #
        loss_weights: np.ndarray = None if 'loss_weights' not in kwargs.keys() else kwargs['loss_weights']
        if dataset:
            if not isinstance(dataset, Dataset):
                raise TypeError(f'{type(dataset)=}    {dataset=}')
            num_classes = num_classes if num_classes is not None else dataset.num_classes
            loss_weights = loss_weights if 'loss_weights' in kwargs.keys() else dataset.loss_weights
        self.num_classes = num_classes  # number of classes
        self.loss_weights = loss_weights  # TODO: what device shall we save loss_weights? numpy, torch, or torch.cuda.
        self.layer_name_list: list[str] = None

        # ------------------------------ #
        self.criterion = self.define_criterion(weight=to_tensor(loss_weights))
        self.criterion_noreduction = self.define_criterion(weight=to_tensor(loss_weights), reduction='none')
        if isinstance(model, type):
            if num_classes is not None:
                kwargs['num_classes'] = num_classes
            self._model = model(name=name, dataset=dataset, **kwargs)
        else:
            assert isinstance(model, nn.Module)
            self._model = model
        self.model = self.get_parallel_model(self._model)
        self.activate_params([])
        if official:
            self.load('official')
        if pretrain:
            self.load(verbose=verbose)
        self.eval()
        if env['num_gpus']:
            self.cuda()

    # ----------------- Forward Operations ----------------------#

    def get_logits(self, _input: torch.Tensor, randomized_smooth: bool = None,
                   rs_sigma: float = None, rs_n: int = None, **kwargs) -> torch.Tensor:
        randomized_smooth = randomized_smooth if randomized_smooth is not None else self.randomized_smooth
        if randomized_smooth:
            rs_sigma = rs_sigma if rs_sigma is not None else self.rs_sigma
            rs_n = rs_n if rs_n is not None else self.rs_n
            _list = []
            for _ in range(rs_n):
                _input_noise = add_noise(_input, std=rs_sigma)  # TODO: valid input clip issue
                _list.append(self.model(_input_noise, **kwargs))
            return torch.stack(_list).mean(dim=0)
            # TODO: memory issues and parallel possibilities
            # _input_noise = add_noise(repeat_to_batch(_input, batch_size=n), std=sigma).flatten(end_dim=1)
            # return self.model(_input_noise, **kwargs).view(n, len(_input), self.num_classes).mean(dim=0)
        else:
            return self.model(_input, **kwargs)

    def get_final_fm(self, _input: torch.Tensor, **kwargs) -> torch.Tensor:
        return self._model.get_final_fm(_input, **kwargs)

    def get_prob(self, _input: torch.Tensor, **kwargs) -> torch.Tensor:
        return self._model.softmax(self(_input, **kwargs))

    def get_target_prob(self, _input: torch.Tensor, target: Union[torch.Tensor, list[int]],
                        **kwargs) -> torch.Tensor:
        if isinstance(target, list):
            target = torch.tensor(target, device=_input.device)
        return self.get_prob(_input, **kwargs).gather(dim=1, index=target.unsqueeze(1)).flatten()

    def get_class(self, _input: torch.Tensor, **kwargs) -> torch.Tensor:
        return self(_input, **kwargs).argmax(dim=-1)

    def get_layer_name(self, depth: int = -1, prefix: str = '',
                       use_filter: bool = True, repeat: bool = False,
                       seq_only: bool = False) -> list[str]:
        return get_layer_name(self._model, depth, prefix, use_filter, repeat, seq_only)

    def get_all_layer(self, x: torch.Tensor,
                      layer_input: str = 'input', depth: int = 0,
                      prefix='', use_filter: bool = True, repeat: bool = False,
                      seq_only: bool = True, verbose: int = 0) -> dict[str, torch.Tensor]:
        return get_all_layer(self._model, x, layer_input, depth,
                             prefix, use_filter, repeat, seq_only, verbose)

    def get_layer(self, x: torch.Tensor, layer_output: str = 'classifier',
                  layer_input: str = 'input', prefix: str = '', seq_only: bool = True) -> torch.Tensor:
        if layer_input == 'input':
            if layer_output == 'classifier':
                return self(x)
            elif layer_output == 'features':
                return self._model.get_fm(x)
            elif layer_output == 'flatten':
                return self.get_final_fm(x)
        if self.layer_name_list is None:
            self.layer_name_list: list[str] = self.get_layer_name(use_filter=False, repeat=True)
            self.layer_name_list.insert(0, 'input')
            self.layer_name_list.append('output')
        return get_layer(self._model, x, layer_output, layer_input, prefix,
                         layer_name_list=self.layer_name_list, seq_only=seq_only)

    def loss(self, _input: torch.Tensor = None, _label: torch.Tensor = None,
             _output: torch.Tensor = None, reduction: str = 'mean',
             **kwargs) -> torch.Tensor:
        criterion = self.criterion_noreduction if reduction == 'none' else self.criterion
        if _output is None:
            _output = self(_input, **kwargs)
        return criterion(_output, _label)

    # -------------------------------------------------------- #

    # Define the optimizer
    # and transfer to that tuning mode.
    # train_opt: 'full' or 'partial' (default: 'partial')
    # lr: (default: [full:2e-3, partial:2e-4])
    # OptimType: to be implemented
    #
    # return: optimizer

    def define_optimizer(self, parameters: Union[str, Iterator[nn.Parameter]] = 'full',
                         OptimType: Union[str, type[Optimizer]] = None,
                         lr: float = 0.1, momentum: float = 0.0, weight_decay: float = 0.0,
                         lr_scheduler: bool = True, T_max: int = None,
                         **kwargs) -> tuple[Optimizer, _LRScheduler]:
        kwargs['momentum'] = momentum
        kwargs['weight_decay'] = weight_decay
        if isinstance(parameters, str):
            parameters = self.get_parameter_from_name(name=parameters)
        if not isinstance(parameters, Iterable):
            raise TypeError(f'{type(parameters)=}    {parameters=}')
        if isinstance(OptimType, str):
            OptimType: type[Optimizer] = getattr(torch.optim, OptimType)
        keys = OptimType.__init__.__code__.co_varnames
        kwargs = {k: v for k, v in kwargs.items() if k in keys}
        optimizer = OptimType(parameters, lr, **kwargs)
        _lr_scheduler: _LRScheduler = None
        if lr_scheduler:
            _lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=T_max)
        return optimizer, _lr_scheduler

    # define loss function
    # Cross Entropy
    # TODO: linting, or maybe nn.Module for generic?
    def define_criterion(self, **kwargs) -> nn.CrossEntropyLoss:
        if 'weight' not in kwargs.keys():
            kwargs['weight'] = self.loss_weights
        return nn.CrossEntropyLoss(**kwargs)
        # if loss_type == 'jsd':
        #     num_classes = num_classes if num_classes is not None else self.num_classes

        #     def jsd(_output: torch.Tensor, _label: torch.Tensor, **kwargs):
        #         p: torch.Tensor = F.one_hot(_label, num_classes)
        #         q: torch.Tensor = F.softmax(_output)
        #         log_q = F.log_softmax(_output)
        #         sum_pq = p + q
        #         loss = sum_pq * (sum_pq.log() - math.log(2))
        # return criterion

    # -----------------------------Load & Save Model------------------------------------------- #

    # file_path: (default: '') if '', use the default path. Else if the path doesn't exist, quit.
    # full: (default: False) whether save feature extractor.
    # output: (default: False) whether output help information.
    def load(self, file_path: str = None, folder_path: str = None, suffix: str = None,
             map_location: Union[str, Callable, torch.device, dict] = 'cpu',
             component: str = '', strict: bool = True,
             verbose: bool = False, indent: int = 0, **kwargs):
        with torch.no_grad():
            map_location = map_location if map_location != 'default' else env['device']
            if file_path is None:
                folder_path = folder_path if folder_path is not None else self.folder_path
                suffix = suffix if suffix is not None else self.suffix
                file_path = os.path.normpath(os.path.join(folder_path, f'{self.name}{suffix}.pth'))
            if file_path == 'official':   # TODO
                _dict = self.get_official_weights(map_location=map_location)
                last_bias_value = next(reversed(_dict.values()))   # TODO: make sure
                if self.num_classes != len(last_bias_value) and component != 'features':
                    strict = False
                    _dict.popitem()
                    _dict.popitem()
            else:
                try:
                    # TODO: type annotation might change? dict[str, torch.Tensor]
                    _dict: OrderedDict[str, torch.Tensor] = torch.load(file_path, map_location=map_location, **kwargs)
                except Exception as e:
                    print(f'{file_path=}')
                    raise e
            module = self._model
            if component == 'features':
                module = self._model.features
                _dict = OrderedDict([(key.removeprefix('features.'), value) for key, value in _dict.items()])
            elif component == 'classifier':
                module = self._model.classifier
                _dict = OrderedDict([(key.removeprefix('classifier.'), value) for key, value in _dict.items()])
            else:
                assert component == '', f'{component=}'
            try:
                module.load_state_dict(_dict, strict=strict)
            except RuntimeError as e:
                prints(f'Model {self.name} loaded from: {file_path}', indent=indent)
                raise e
            if verbose:
                prints(f'Model {self.name} loaded from: {file_path}', indent=indent)
            if env['num_gpus']:
                self.cuda()

    # file_path: (default: '') if '', use the default path.
    # full: (default: False) whether save feature extractor.
    def save(self, file_path: str = None, folder_path: str = None, suffix: str = None,
             component: str = '', verbose: bool = False, indent: int = 0, **kwargs):
        with torch.no_grad():
            if file_path is None:
                folder_path = folder_path if folder_path is not None else self.folder_path
                suffix = suffix if suffix is not None else self.suffix
                file_path = os.path.normpath(os.path.join(folder_path, f'{self.name}{suffix}.pth'))
            else:
                folder_path = os.path.dirname(file_path)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
            # TODO: type annotation might change? dict[str, torch.Tensor]
            module = self._model
            if component == 'features':
                module = self._model.features
            elif component == 'classifier':
                module = self._model.classifier
            else:
                assert component == '', f'{component=}'
            _dict: OrderedDict[str, torch.Tensor] = module.state_dict(prefix=component)
            torch.save(_dict, file_path, **kwargs)
            if verbose:
                prints(f'Model {self.name} saved at: {file_path}', indent=indent)

    def get_official_weights(self, url: str = None,
                             map_location: Union[str, Callable, torch.device, dict] = 'cpu',
                             **kwargs) -> OrderedDict[str, torch.Tensor]:
        url = self.model_urls[self.name] if url is None else url
        print('get official model weights from: ', url)
        return model_zoo.load_url(url, map_location=map_location, **kwargs)

    # -----------------------------------Train and Validate------------------------------------ #
    # TODO: annotation and remove those arguments to be *args, **kwargs
    def _train(self, epoch: int, optimizer: Optimizer, lr_scheduler: _LRScheduler = None,
               grad_clip: float = None, pre_conditioner: Union[KFAC, EKFAC] = None,
               print_prefix: str = 'Epoch', start_epoch: int = 0, resume: int = 0,
               validate_interval: int = 10, save: bool = False, amp: bool = False,
               loader_train: torch.utils.data.DataLoader = None, loader_valid: torch.utils.data.DataLoader = None,
               epoch_fn: Callable[..., None] = None,
               get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
               loss_fn: Callable[..., torch.Tensor] = None,
               after_loss_fn: Callable[..., None] = None,
               validate_fn: Callable[..., tuple[float, float]] = None,
               save_fn: Callable[..., None] = None, file_path: str = None, folder_path: str = None, suffix: str = None,
               writer=None, main_tag: str = 'train', tag: str = '',
               accuracy_fn: Callable[..., list[float]] = None,
               verbose: bool = True, indent: int = 0, **kwargs) -> None:
        loader_train = loader_train if loader_train is not None else self.dataset.loader['train']
        get_data_fn = get_data_fn if callable(get_data_fn) else self.get_data
        loss_fn = loss_fn if callable(loss_fn) else self.loss
        validate_fn = validate_fn if callable(validate_fn) else self._validate
        save_fn = save_fn if callable(save_fn) else self.save
        accuracy_fn = accuracy_fn if callable(accuracy_fn) else self.accuracy
        # if not callable(iter_fn) and hasattr(self, 'iter_fn'):
        #     iter_fn = getattr(self, 'iter_fn')
        if not callable(epoch_fn) and hasattr(self, 'epoch_fn'):
            epoch_fn = getattr(self, 'epoch_fn')
        if not callable(after_loss_fn) and hasattr(self, 'after_loss_fn'):
            after_loss_fn = getattr(self, 'after_loss_fn')
        return train(module=self, num_classes=self.num_classes,
                     epoch=epoch, optimizer=optimizer, lr_scheduler=lr_scheduler,
                     grad_clip=grad_clip, pre_conditioner=pre_conditioner,
                     print_prefix=print_prefix, start_epoch=start_epoch, resume=resume,
                     validate_interval=validate_interval, save=save, amp=amp,
                     loader_train=loader_train, loader_valid=loader_valid,
                     epoch_fn=epoch_fn, get_data_fn=get_data_fn, loss_fn=loss_fn, after_loss_fn=after_loss_fn, validate_fn=validate_fn,
                     save_fn=save_fn, file_path=file_path, folder_path=folder_path, suffix=suffix,
                     writer=writer, main_tag=main_tag, tag=tag,
                     accuracy_fn=accuracy_fn, verbose=verbose, indent=indent, **kwargs)

    def _validate(self, module: nn.Module = None, num_classes: int = None,
                  full: bool = True, loader: torch.utils.data.DataLoader = None,
                  print_prefix: str = 'Validate', indent: int = 0, verbose: bool = True,
                  get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                  loss_fn: Callable[..., torch.Tensor] = None,
                  writer=None, main_tag: str = 'valid', tag: str = '', _epoch: int = None,
                  accuracy_fn: Callable[..., list[float]] = None,
                  **kwargs) -> tuple[float, float]:
        module = self if module is None else module
        num_classes = self.num_classes if num_classes is None else num_classes
        if loader is None:
            loader = self.dataset.loader['valid'] if full else self.dataset.loader['valid2']
        get_data_fn = get_data_fn if get_data_fn is not None else self.get_data
        loss_fn = loss_fn if loss_fn is not None else self.loss
        accuracy_fn = accuracy_fn if callable(accuracy_fn) else self.accuracy
        return validate(module=module, num_classes=num_classes, loader=loader,
                        print_prefix=print_prefix, indent=indent, verbose=verbose,
                        get_data_fn=get_data_fn, loss_fn=loss_fn,
                        writer=writer, main_tag=main_tag, tag=tag,
                        _epoch=_epoch, accuracy_fn=accuracy_fn, **kwargs)

    # TODO: this method shall be removed
    def _compare(self, peer: nn.Module = None, full: bool = True, loader: torch.utils.data.DataLoader = None,
                 print_prefix: str = 'Validate', indent: int = 0, verbose: bool = True,
                 get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                 **kwargs) -> tuple[float, float]:
        import warnings
        warnings.warn('This method shall be removed. You should call `autovul.vision.utils.train.compare` directly.',
                      DeprecationWarning)
        module1 = self  # TODO: type annotation issues (solve in python 3.10)
        module2 = peer
        if loader is None:
            loader = self.dataset.loader['valid'] if full else self.dataset.loader['valid2']
        get_data_fn = get_data_fn if get_data_fn is not None else self.get_data
        return compare(module1, module2, loader,
                       print_prefix, indent, verbose,
                       get_data_fn, **kwargs)

    # -------------------------------------------Utility--------------------------------------- #

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        if self.dataset is not None:
            return self.dataset.get_data(data, **kwargs)
        else:
            return data

    def accuracy(self, _output: torch.Tensor, _label: torch.Tensor, num_classes: int = None,
                 topk: tuple[int] = (1, 5)) -> list[float]:
        num_classes = num_classes if num_classes is not None else self.num_classes
        return accuracy(_output, _label, num_classes, topk)

    def get_parameter_from_name(self, name: str = '') -> Iterator[nn.Parameter]:
        params = self._model.parameters()
        if name == 'features':
            params = self._model.features.parameters()
        elif name in ['classifier', 'partial']:
            params = self._model.classifier.parameters()
        elif name not in ['', 'full']:
            raise NotImplementedError(f'{name=}')
        return params

    def activate_params(self, params: Iterator[nn.Parameter]) -> None:
        return activate_params(self._model, params)

    # Need to overload for other packages (GNN) since they are calling their own nn.DataParallel.
    # TODO: nn.parallel.DistributedDataParallel
    @staticmethod
    def get_parallel_model(_model: _Model) -> Union[_Model, nn.DataParallel]:
        if env['num_gpus'] > 1:
            return nn.DataParallel(_model)
        return _model

    def summary(self, depth: int = None, verbose: bool = True, indent: int = 0, **kwargs):
        if depth is None:
            depth = env['verbose']
        if depth is None:
            depth = 1
        prints('{blue_light}{0:<30s}{reset} Parameters: '.format(self.name, **ansi), indent=indent)
        prints(self.__class__.__name__, indent=indent)
        for key, value in self.param_list.items():
            if value:
                prints('{green}{0:<20s}{reset}'.format(key, **ansi), indent=indent + 10)
                prints({v: getattr(self, v) for v in value}, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)
        summary(self._model, depth=depth, verbose=verbose, indent=indent + 10, **kwargs)
        prints('-' * 20, indent=indent + 10)

    # -----------------------------------------Reload------------------------------------------ #

    def __call__(self, _input: torch.Tensor, amp: bool = False, **kwargs) -> torch.Tensor:
        if amp:
            with torch.cuda.amp.autocast():
                return self.get_logits(_input, **kwargs)
        return self.get_logits(_input, **kwargs)

    # def __str__(self) -> str:
    #     sys.stdout = redirect
    #     self.summary()
    #     _str = redirect.buffer
    #     redirect.reset()
    #     return _str

    # def __repr__(self):
    #     return self.name

    def train(self, mode: bool = True):
        self._model.train(mode=mode)
        self.model.train(mode=mode)
        return self

    def eval(self):
        self._model.eval()
        self.model.eval()
        return self

    def cpu(self):
        self._model.cpu()
        self.model.cpu()
        return self

    def cuda(self, device: Optional[Union[int, torch.device]] = None):
        self._model.cuda(device=device)
        self.model.cuda(device=device)
        return self

    def zero_grad(self, set_to_none: bool = False):
        return self._model.zero_grad(set_to_none=set_to_none)

    def state_dict(self, destination: Mapping[str, torch.Tensor] = None, prefix: str = '', keep_vars: bool = False):
        return self._model.state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)

    def load_state_dict(self, state_dict: dict[str, torch.Tensor], strict: bool = True):
        return self._model.load_state_dict(state_dict, strict=strict)

    def parameters(self, recurse: bool = True):
        return self._model.parameters(recurse=recurse)

    def named_parameters(self, prefix='', recurse=True):
        return self._model.named_parameters(prefix=prefix, recurse=recurse)

    def children(self):
        return self._model.children()

    def named_children(self):
        return self._model.named_children()

    def modules(self):
        return self._model.modules()

    # TODO: why use the '' for 'nn.Module'?
    def named_modules(self, memo: Optional[Set['nn.Module']] = None, prefix: str = '') -> Generator[tuple[str, nn.Module], None, None]:
        return self._model.named_modules(memo=memo, prefix=prefix)

    def apply(self, fn: Callable[['nn.Module'], None]):
        return self._model.apply(fn)

    # ----------------------------------------------------------------------------------------- #

    def remove_misclassify(self, data: tuple[torch.Tensor, torch.Tensor], **kwargs):
        with torch.no_grad():
            _input, _label = self.get_data(data, **kwargs)
            _classification = self.get_class(_input)
            repeat_idx = _classification.eq(_label)
        return _input[repeat_idx], _label[repeat_idx]

    def generate_target(self, _input: torch.Tensor, idx: int = 1, same: bool = False) -> torch.Tensor:
        return generate_target(self, _input, idx, same)


def add_argument(parser: argparse.ArgumentParser, model_name: str = None, model: Union[str, Model] = None,
                 config: Config = config, class_dict: dict[str, type[Model]] = {}):
    dataset_name = get_name(arg_list=['-d', '--dataset'])
    if dataset_name is None:
        dataset_name = config.get_full_config()['dataset']['default_dataset']
    model_name = get_name(name=model_name, module=model, arg_list=['-m', '--model'])
    if model_name is None:
        model_name = config.get_config(dataset_name=dataset_name)['model']['default_model']
    model_name = get_model_class(model_name, class_dict=class_dict)

    group = parser.add_argument_group('{yellow}model{reset}'.format(**ansi), description=model_name)
    model_class_name = get_model_class(model_name, class_dict=class_dict)
    try:
        ModelType = class_dict[model_class_name]
    except KeyError as e:
        print(f'{model_class_name} not in \n{list(class_dict.keys())}')
        raise e
    return ModelType.add_argument(group)


def create(model_name: str = None, model: Union[str, Model] = None,
           dataset_name: str = None, dataset: Union[str, Dataset] = None,
           folder_path: str = None,
           config: Config = config, class_dict: dict[str, type[Model]] = {}, **kwargs) -> Model:
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    model_name = get_name(name=model_name, module=model, arg_list=['-m', '--model'])
    if dataset_name is None:
        dataset_name = config.get_full_config()['dataset']['default_dataset']
    if model_name is None:
        model_name = config.get_config(dataset_name=dataset_name)['model']['default_model']
    result = config.get_config(dataset_name=dataset_name)['model'].update(kwargs)
    model_name = model_name if model_name is not None else result['default_model']

    name_list = [name for sub_list in get_available_models(class_dict=class_dict).values()
                 for name in sub_list]
    name_list = sorted(name_list)
    assert model_name in name_list, f'{model_name} not in \n{name_list}'
    model_class_name = get_model_class(model_name, class_dict=class_dict)
    try:
        ModelType = class_dict[model_class_name]
    except KeyError as e:
        print(f'{model_class_name} not in \n{list(class_dict.keys())}')
        raise e
    if folder_path is None and isinstance(dataset, Dataset):
        folder_path = os.path.join(result['model_dir'], dataset.data_type, dataset.name)
    return ModelType(name=model_name, dataset=dataset, folder_path=folder_path, **result)


def get_available_models(class_dict: dict[str, type[Model]] = {}) -> dict[str, list[str]]:
    return {k: v.available_models for k, v in class_dict.items()}


def output_available_models(class_dict: dict[str, type[Model]] = {}, indent: int = 0) -> None:
    names_dict = get_available_models(class_dict)
    for k in sorted(names_dict.keys()):
        prints('{yellow}{k}{reset}'.format(k=k, **ansi), indent=indent)
        prints(names_dict[k], indent=indent + 10)
        print()


def get_model_class(name: str, class_dict: dict[str, type[Model]] = {}) -> str:
    correct_name: str = None
    for class_name in class_dict.keys():
        if class_name in name.lower() \
                and (correct_name is None or len(class_name) > len(correct_name)):
            correct_name = class_name
    if correct_name is not None:
        return correct_name
    raise KeyError(f'{name} not in {list(class_dict.keys())}')



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch
import torch.hub
import torch.nn as nn

import argparse
from collections import OrderedDict


class _ProxylessNAS(_ImageModel):

    def __init__(self, target_platform: str = 'proxyless_cifar', **kwargs):
        super().__init__(**kwargs)
        _model = torch.hub.load('ain-soph/ProxylessNAS', target_platform, pretrained=False)
        self.features = nn.Sequential()
        if _model.__class__.__name__ == 'ProxylessNASNets':
            self.features.add_module('first_conv', _model.first_conv)
            self.features.add_module('blocks', nn.Sequential(*_model.blocks))
            if hasattr(_model, 'feature_mix_layer'):
                self.features.add_module('feature_mix_layer', _model.feature_mix_layer)
        else:
            assert _model.__class__.__name__ == 'PyramidTreeNet', _model.__class__.__name__
            self.features.add_module('blocks', nn.Sequential(*_model.blocks[:-1]))
        classifier: nn.Module = _model.classifier
        fc: nn.Linear = list(classifier.children())[0]
        self.classifier = self.define_classifier(conv_dim=fc.in_features,
                                                 num_classes=self.num_classes,
                                                 fc_depth=1)


class ProxylessNAS(ImageModel):
    available_models = ['proxylessnas']

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--target_platform')
        return group

    def __init__(self, name: str = 'proxylessnas', target_platform: str = 'proxyless_cifar',
                 model: type[_ProxylessNAS] = _ProxylessNAS, **kwargs):
        self.target_platform = target_platform
        super().__init__(name=name, model=model, target_platform=target_platform, **kwargs)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        _model: nn.Module = torch.hub.load('ain-soph/ProxylessNAS', self.target_platform, pretrained=True)
        _dict = OrderedDict()
        if hasattr(_model, 'first_conv'):
            first_conv: nn.Module = _model.first_conv
            _dict.update(first_conv.state_dict(prefix='features.first_conv.'))
        blocks: nn.ModuleList = _model.blocks
        if self.target_platform == 'proxyless_cifar':
            blocks = blocks[:-1]
        _dict.update(blocks.state_dict(prefix='features.blocks.'))
        if hasattr(_model, 'feature_mix_layer'):
            feature_mix_layer: nn.Module = _model.feature_mix_layer
            _dict.update(feature_mix_layer.state_dict(prefix='features.feature_mix_layer.'))
        fc: nn.Linear = list(_model.classifier.children())[0]
        _dict.update(fc.state_dict(prefix='classifier.fc.'))
        return _dict
